from __future__ import print_function

import itertools
import gevent.socket
import json
import os
import sys
import uuid
import socket

import fastremap
from six.moves import range
import numpy as np
from tqdm import tqdm
from six import string_types
import multiprocessing as mp

from .. import lib
from ..cacheservice import CacheService
from .. import exceptions 
from ..lib import ( 
  colorize, red, mkdir, 
  Vec, Bbox, jsonify
)

from ..datasource import autocropfn
from ..datasource.precomputed import PrecomputedMetadata

from ..paths import strict_extract
from ..provenance import DataLayerProvenance
from ..storage import SimpleStorage, Storage, reset_connection_pools
from ..volumecutout import VolumeCutout
from .. import sharedmemory

def warn(text):
  print(colorize('yellow', text))

class CloudVolumePrecomputed(object):
  def __init__(self, 
    meta, cache, config,
    image=None, mesh=None, skeleton=None,
    mip=0
  ):
    self.config = config 
    self.cache = cache 
    self.meta = meta

    self.image = image
    self.mesh = mesh 
    self.skeleton = skeleton

    self.green_threads = self.config.green # display warning message

    # needs to be set after info is defined since
    # its setter is based off of scales
    self.mip = mip
    self.pid = os.getpid()

  @property 
  def autocrop(self):
    return self.image.autocrop

  @autocrop.setter
  def autocrop(self, val):
    self.image.autocrop = val

  @property
  def background_color(self):
    return self.image.background_color 

  @background_color.setter
  def background_color(self, val):
    self.image.background_color = val  

  @property 
  def bounded(self):
    return self.image.bounded

  @bounded.setter 
  def bounded(self, val):
    self.image.bounded = val

  @property
  def fill_missing(self):
    return self.image.fill_missing
  
  @fill_missing.setter
  def fill_missing(self, val):
    self.image.fill_missing = val
  
  @property
  def green_threads(self):
    return self.config.green
  
  @green_threads.setter 
  def green_threads(self, val):
    if val and socket.socket is not gevent.socket.socket:
      warn("""
      WARNING: green_threads is set but this process is
      not monkey patched. This will cause severely degraded
      performance.
      
      CloudVolume uses gevent for cooperative (green)
      threading but it requires patching the Python standard
      library to perform asynchronous IO. Add this code to
      the top of your program (before any other imports):

        import gevent.monkey
        gevent.monkey.patch_all(threads=False)

      More Information:

      http://www.gevent.org/intro.html#monkey-patching
      """)

    self.config.green = bool(val)

  @property
  def non_aligned_writes(self):
    return self.image.non_aligned_writes

  @non_aligned_writes.setter
  def non_aligned_writes(self, val):
    self.image.non_aligned_writes = val

  @property
  def delete_black_uploads(self):
    return self.image.delete_black_uploads

  @delete_black_uploads.setter
  def delete_black_uploads(self, val):
    self.image.delete_black_uploads = val

  @property
  def parallel(self):
    return self.config.parallel

  @parallel.setter
  def parallel(self, num_processes):
    if type(num_processes) == bool:
      num_processes = mp.cpu_count() if num_processes == True else 1
    elif num_processes <= 0:
      raise ValueError('Number of processes must be >= 1. Got: ' + str(num_processes))
    else:
      num_processes = int(num_processes)

    self.config.parallel = num_processes

  @property
  def cdn_cache(self):
    return self.config.cdn_cache

  @cdn_cache.setter 
  def cdn_cache(self, val):
    self.config.cdn_cache = val

  @property 
  def compress(self):
    return self.config.compress

  @compress.setter 
  def compress(self, val):
    self.config.compress = val 

  @property 
  def progress(self):
    return self.config.progress 

  @progress.setter 
  def progress(self, val):
    self.config.progress = bool(val)

  @property 
  def info(self):
    return self.meta.info

  @info.setter
  def info(self, val):
    self.meta.info = val
  
  @property
  def provenance(self):
    return self.meta.provenance

  @provenance.setter
  def provenance(self, val):
    self.meta.provenance = val

  def __setstate__(self, d):
    """Called when unpickling which is integral to multiprocessing."""
    self.__dict__ = d 
    
    pid = os.getpid()
    if 'pid' in d and d['pid'] != pid:
      # otherwise the pickle might have references to old connections
      reset_connection_pools() 
      self.pid = pid
  
  @classmethod
  def create_new_info(cls, 
    num_channels, layer_type, data_type, encoding, 
    resolution, voxel_offset, volume_size, 
    mesh=None, skeletons=None, chunk_size=(64,64,64),
    compressed_segmentation_block_size=(8,8,8),
    max_mip=0, factor=Vec(2,2,1), redirect=None, 
    *args, **kwargs
  ):
    """
    Create a new neuroglancer Precomputed info file.

    Required:
      num_channels: (int) 1 for grayscale, 3 for RGB 
      layer_type: (str) typically "image" or "segmentation"
      data_type: (str) e.g. "uint8", "uint16", "uint32", "float32"
      encoding: (str) "raw" for binaries like numpy arrays, "jpeg"
      resolution: int (x,y,z), x,y,z voxel dimensions in nanometers
      voxel_offset: int (x,y,z), beginning of dataset in positive cartesian space
      volume_size: int (x,y,z), extent of dataset in cartesian space from voxel_offset
    
    Optional:
      mesh: (str) name of mesh directory, typically "mesh"
      skeletons: (str) name of skeletons directory, typically "skeletons"
      chunk_size: int (x,y,z), dimensions of each downloadable 3D image chunk in voxels
      compressed_segmentation_block_size: (x,y,z) dimensions of each compressed sub-block
        (only used when encoding is 'compressed_segmentation')
      max_mip: (int), the maximum mip level id.
      factor: (Vec), the downsampling factor for each mip level
      redirect: If this volume has moved, you can set an automatic redirect
        by specifying a cloudpath here.

    Returns: dict representing a single mip level that's JSON encodable
    """
    return PrecomputedMetadata.create_info(
      num_channels, layer_type, data_type, encoding, 
      resolution, voxel_offset, volume_size, 
      mesh, skeletons, chunk_size,
      compressed_segmentation_block_size,
      max_mip, factor,
      *args, **kwargs
    )

  def refresh_info(self):
    """Restore the current info from cache or storage."""
    return self.meta.refresh_info()

  def commit_info(self):
    return self.meta.commit_info()

  def refresh_provenance(self):
    return self.meta.refresh_provenance()

  def commit_provenance(self):
    return self.meta.commit_provenance()

  @property
  def dataset_name(self):
    return self.meta.dataset
  
  @property
  def layer(self):
    return self.meta.layer

  @property
  def mip(self):
    return self.config.mip

  @mip.setter
  def mip(self, mip):
    self.config.mip = self.meta.to_mip(mip)

  @property
  def scales(self):
    return self.meta.scales

  @scales.setter
  def scales(self, val):
    self.meta.scales = val

  @property
  def scale(self):
    return self.meta.scale(self.mip)

  @scale.setter
  def scale(self, val):
    self.info['scales'][self.mip] = val

  def mip_scale(self, mip):
    return self.meta.scale(mip)

  @property
  def basepath(self):
    return self.meta.basepath

  @property 
  def layerpath(self):
    return self.meta.layerpath

  @property
  def base_cloudpath(self):
    return self.meta.base_cloudpath

  @property 
  def cloudpath(self):
    return self.layer_cloudpath

  @property
  def layer_cloudpath(self):
    return self.meta.cloudpath

  @property
  def info_cloudpath(self):
    return self.meta.infopath

  @property
  def cache_path(self):
    return self.cache.path

  @property
  def ndim(self):
    return len(self.shape)

  def mip_ndim(self, mip):
    return len(self.meta.shape(mip))

  @property
  def shape(self):
    """Returns Vec(x,y,z,channels) shape of the volume similar to numpy.""" 
    return tuple(self.meta.shape(self.mip))

  def mip_shape(self, mip):
    return tuple(self.meta.shape(mip))

  @property
  def volume_size(self):
    """Returns Vec(x,y,z) shape of the volume (i.e. shape - channels).""" 
    return self.meta.volume_size(self.mip)

  def mip_volume_size(self, mip):
    return self.meta.volume_size(mip)

  @property
  def available_mips(self):
    """Returns a list of mip levels that are defined."""
    return self.meta.available_mips

  @property
  def available_resolutions(self):
    """Returns a list of defined resolutions."""
    return (s["resolution"] for s in self.scales)

  @property
  def layer_type(self):
    """e.g. 'image' or 'segmentation'"""
    return self.meta.layer_type

  @property
  def dtype(self):
    """e.g. 'uint8'"""
    return self.meta.dtype

  @property
  def data_type(self):
    return self.meta.data_type

  @property
  def encoding(self):
    """e.g. 'raw' or 'jpeg'"""
    return self.meta.encoding(self.mip)

  def mip_encoding(self, mip):
    return self.meta.encoding(mip)

  @property
  def compressed_segmentation_block_size(self):
    return self.mip_compressed_segmentation_block_size(self.mip)

  def mip_compressed_segmentation_block_size(self, mip):
    if 'compressed_segmentation_block_size' in self.info['scales'][mip]:
      return self.info['scales'][mip]['compressed_segmentation_block_size']
    return None

  @property
  def num_channels(self):
    return self.meta.num_channels

  @property
  def voxel_offset(self):
    """Vec(x,y,z) start of the dataset in voxels"""
    return self.meta.voxel_offset(self.mip)

  def mip_voxel_offset(self, mip):
    return self.meta.voxel_offset(mip)

  @property 
  def resolution(self):
    """Vec(x,y,z) dimensions of each voxel in nanometers"""
    return self.meta.resolution(self.mip)

  def mip_resolution(self, mip):
    return self.meta.resolution(mip)

  @property
  def downsample_ratio(self):
    """Describes how downsampled the current mip level is as an (x,y,z) factor triple."""
    return self.meta.downsample_ratio(self.mip)

  @property
  def chunk_size(self):
    """Underlying chunk size dimensions in voxels. Synonym for underlying."""
    return self.meta.chunk_size(self.mip)

  def mip_chunk_size(self, mip):
    return self.meta.chunk_size(mip)

  @property
  def underlying(self):
    """Underlying chunk size dimensions in voxels. Synonym for chunk_size."""
    return self.meta.chunk_size(self.mip)

  def mip_underlying(self, mip):
    return self.meta.chunk_size(mip)

  @property
  def key(self):
    """The subdirectory within the data layer containing the chunks for this mip level"""
    return self.meta.key(self.mip)

  def mip_key(self, mip):
    return self.meta.key(mip)

  @property
  def bounds(self):
    """Returns a bounding box for the dataset with dimensions in voxels"""
    return self.meta.bounds(self.mip)

  def mip_bounds(self, mip):
    offset = self.meta.voxel_offset(mip)
    shape = self.meta.volume_size(mip)
    return Bbox( offset, offset + shape )

  def point_to_mip(self, pt, mip, to_mip):
    return self.meta.point_to_mip(pt, mip, to_mip)

  def bbox_to_mip(self, bbox, mip, to_mip):
    """Convert bbox or slices from one mip level to another."""
    return self.meta.bbox_to_mip(bbox, mip, to_mip)

  def slices_to_global_coords(self, slices):
    """
    Used to convert from a higher mip level into mip 0 resolution.
    """
    bbox = self.meta.bbox_to_mip(slices, self.mip, 0)
    return bbox.to_slices()

  def slices_from_global_coords(self, slices):
    """
    Used for converting from mip 0 coordinates to upper mip level
    coordinates. This is mainly useful for debugging since the neuroglancer
    client displays the mip 0 coordinates for your cursor.
    """
    bbox = self.meta.bbox_to_mip(slices, 0, self.mip)
    return bbox.to_slices()

  def reset_scales(self):
    """Used for manually resetting downsamples if something messed up."""
    self.meta.reset_scales()
    return self.commit_info()

  def add_scale(self, factor, encoding=None, chunk_size=None, info=None):
    """
    Generate a new downsample scale to for the info file and return an updated dictionary.
    You'll still need to call self.commit_info() to make it permenant.

    Required:
      factor: int (x,y,z), e.g. (2,2,1) would represent a reduction of 2x in x and y

    Optional:
      encoding: force new layer to e.g. jpeg or compressed_segmentation
      chunk_size: force new layer to new chunk size

    Returns: info dict
    """
    return self.meta.add_scale(factor, encoding, chunk_size, info)

  def exists(self, bbox_or_slices):
    """
    Produce a summary of whether all the requested chunks exist.

    bbox_or_slices: accepts either a Bbox or a tuple of slices representing
      the requested volume.
    Returns: { chunk_file_name: boolean, ... }
    """
    return self.image.exists(bbox_or_slices)

  def delete(self, bbox_or_slices):
    """
    Delete the files within the bounding box.

    bbox_or_slices: accepts either a Bbox or a tuple of slices representing
      the requested volume. 
    """
    return self.image.delete(bbox_or_slices)

  def transfer_to(self, cloudpath, bbox, block_size=None, compress=True):
    """
    Transfer files from one storage location to another, bypassing
    volume painting. This enables using a single CloudVolume instance
    to transfer big volumes. In some cases, gsutil or aws s3 cli tools
    may be more appropriate. This method is provided for convenience. It
    may be optimized for better performance over time as demand requires.

    cloudpath (str): path to storage layer
    bbox (Bbox object): ROI to transfer
    block_size (int): number of file chunks to transfer per I/O batch.
    compress (bool): Set to False to upload as uncompressed
    """
    return self.image.transfer_to(cloudpath, bbox, self.mip, block_size, compress)

  def __getitem__(self, slices):
    if type(slices) == Bbox:
      slices = slices.to_slices()

    slices = self.meta.bbox(self.mip).reify_slices(slices, bounded=self.bounded)
    steps = Vec(*[ slc.step for slc in slices ])
    channel_slice = slices.pop()
    requested_bbox = Bbox.from_slices(slices)

    img = self.download(requested_bbox, self.mip)
    return img[::steps.x, ::steps.y, ::steps.z, channel_slice]

  def download(
      self, bbox, mip=None, parallel=None,
      segids=None, preserve_zeros=False,
      
      # Absorbing polymorphic Graphene calls
      agglomerate=None, timestamp=None, stop_layer=None
    ):
    """
    Downloads segmentation from the indicated cutout
    region.

    bbox: specifies cutout to fetch
    mip: which resolution level to get (default self.mip)
    parallel: what parallel level to use (default self.parallel)

    segids: agglomerate the leaves of these segids from the graph 
      server and label them with the given segid.
    preserve_zeros: If segids is not None:
      False: mask other segids with zero
      True: mask other segids with the largest integer value
        contained by the image data type and leave zero as is.

    agglomerate, timestamp, and stop_layer are just there to 
    absorb arguments to what could be a graphene frontend.

    Returns: img
    """  
    bbox = Bbox.create(
      bbox, context=self.bounds, 
      bounded=self.bounded, 
      autocrop=self.autocrop
    )

    if mip is None:
      mip = self.mip

    if parallel is None:
      parallel = self.parallel

    img = self.image.download(bbox, mip, parallel=parallel)

    if segids is None:
      return img

    mask_value = 0
    if preserve_zeros:
      mask_value = np.inf
      if np.issubdtype(self.dtype, np.integer):
        mask_value = np.iinfo(self.dtype).max

      segids.append(0)

    img = fastremap.mask_except(img, segids, in_place=True, value=mask_value)

    return VolumeCutout.from_volume(
      self.meta, mip, img, bbox
    )

  def download_point(
    self, pt, size=256, 
    mip=None, parallel=None, 
    coord_resolution=None,
    **kwargs
  ):
    """
    Download to the right of point given in mip 0 coords.
    Useful for quickly visualizing a neuroglancer coordinate
    at an arbitary mip level.

    pt: (x,y,z)
    size: int or (sx,sy,sz)
    mip: int representing resolution level
    parallel: number of processes to launch (0 means all cores)
    coord_resolution: (rx,ry,rz) the coordinate resolution of the input point.
      Sometimes Neuroglancer is working in the resolution of another
      higher res layer and this can help correct that.

    Return: image
    """
    if isinstance(size, int):
      size = Vec(size, size, size)
    else:
      size = Vec(*size)

    if mip is None:
      mip = self.mip

    mip = self.meta.to_mip(mip)
    size2 = size // 2

    if coord_resolution is not None:
      factor = self.meta.resolution(0) / Vec(*coord_resolution)
      pt = Vec(*pt) / factor

    pt = self.point_to_mip(pt, mip=0, to_mip=mip)
    bbox = Bbox(pt - size2, pt + size2).astype(np.int64)
    for i, sz in enumerate(size):
      if sz == 1:
        bbox.minpt[i] = pt[i]
        bbox.maxpt[i] = pt[i] + 1

    if self.autocrop:
      bbox = Bbox.intersection(bbox, self.meta.bounds(mip))

    bbox = bbox.astype(np.int32)
    if parallel is None:
      parallel = self.parallel

    return self.image.download(bbox, mip, parallel=parallel, **kwargs)

  def unlink_shared_memory(self):
    """Unlink the current shared memory location from the filesystem."""
    return self.image.unlink_shared_memory()

  def download_to_shared_memory(self, slices, location=None, mip=None):
    """
    Download images to a shared memory array. 

    https://github.com/seung-lab/cloud-volume/wiki/Advanced-Topic:-Shared-Memory

    tip: If you want to use slice notation, np.s_[...] will help in a pinch.

    MEMORY LIFECYCLE WARNING: You are responsible for managing the lifecycle of the 
      shared memory. CloudVolume will merely write to it, it will not unlink the 
      memory automatically. To fully clear the shared memory you must unlink the 
      location and close any mmap file handles. You can use `cloudvolume.sharedmemory.unlink(...)`
      to help you unlink the shared memory file or `vol.unlink_shared_memory()` if you do 
      not specify location (meaning the default instance location is used).

    EXPERT MODE WARNING: If you aren't sure you need this function (e.g. to relieve 
      memory pressure or improve performance in some way) you should use the ordinary 
      download method of img = vol[:]. A typical use case is transferring arrays between 
      different processes without making copies. For reference, this  feature was created 
      for downloading a 62 GB array and working with it in Julia.

    Required:
      slices: (Bbox or list of slices) the bounding box the shared array represents. For instance
        if you have a 1024x1024x128 volume and you're uploading only a 512x512x64 corner
        touching the origin, your Bbox would be `Bbox( (0,0,0), (512,512,64) )`.
    Optional:
      location: (str) Defaults to self.shared_memory_id. Shared memory location 
        e.g. 'cloudvolume-shm-RANDOM-STRING' This typically corresponds to a file 
        in `/dev/shm` or `/run/shm/`. It can also be a file if you're using that for mmap. 
    
    Returns: ndarray backed by shared memory
    """
    if mip is None:
      mip = self.mip

    slices = self.meta.bbox(mip).reify_slices(slices, bounded=self.bounded)
    steps = Vec(*[ slc.step for slc in slices ])
    channel_slice = slices.pop()
    requested_bbox = Bbox.from_slices(slices)

    if self.autocrop:
      requested_bbox = Bbox.intersection(requested_bbox, self.bounds)

    img = self.image.download(
      requested_bbox, mip, parallel=self.parallel,
      location=location, retain=True, use_shared_memory=True
    )
    return img[::steps.x, ::steps.y, ::steps.z, channel_slice]

  def download_to_file(self, path, bbox, mip=None):
    """
    Download images directly to a file.

    Required:
      slices: (Bbox) the bounding box the shared array represents. For instance
        if you have a 1024x1024x128 volume and you're uploading only a 512x512x64 corner
        touching the origin, your Bbox would be `Bbox( (0,0,0), (512,512,64) )`.
      path: (str)
    Optional:
      mip: (int; default: self.mip) The current resolution level.

    Returns: ndarray backed by an mmapped file
    """
    if mip is None:
      mip = self.mip

    slices = self.meta.bbox(mip).reify_slices(bbox, bounded=self.bounded)
    steps = Vec(*[ slc.step for slc in slices ])
    channel_slice = slices.pop()
    requested_bbox = Bbox.from_slices(slices)

    if self.autocrop:
      requested_bbox = Bbox.intersection(requested_bbox, self.bounds)

    img = self.image.download(
      requested_bbox, mip, parallel=self.parallel,
      location=lib.toabs(path), retain=True, use_file=True
    )
    return img[::steps.x, ::steps.y, ::steps.z, channel_slice]

  def __setitem__(self, slices, img):
    if type(slices) == Bbox:
      slices = slices.to_slices()

    slices = self.meta.bbox(self.mip).reify_slices(slices, bounded=self.bounded)
    bbox = Bbox.from_slices(slices)
    slice_shape = list(bbox.size())
    bbox = Bbox.from_slices(slices[:3])

    if np.isscalar(img):
      img = np.zeros(slice_shape, dtype=self.dtype) + img

    imgshape = list(img.shape)
    if len(imgshape) == 3:
      imgshape = imgshape + [ self.num_channels ]

    if not np.array_equal(imgshape, slice_shape):
      raise exceptions.AlignmentError("""
        Input image shape does not match slice shape.

        Image Shape: {}  
        Slice Shape: {}
      """.format(imgshape, slice_shape))

    if self.autocrop:
      if not self.bounds.contains_bbox(bbox):
        img, bbox = autocropfn(self.meta, img, bbox, self.mip)

    if bbox.subvoxel():
      return

    self.image.upload(img, bbox.minpt, self.mip, parallel=self.parallel)

  def upload_from_shared_memory(self, location, bbox, order='F', cutout_bbox=None):
    """
    Upload from a shared memory array. 

    https://github.com/seung-lab/cloud-volume/wiki/Advanced-Topic:-Shared-Memory

    tip: If you want to use slice notation, np.s_[...] will help in a pinch.

    MEMORY LIFECYCLE WARNING: You are responsible for managing the lifecycle of the 
      shared memory. CloudVolume will merely read from it, it will not unlink the 
      memory automatically. To fully clear the shared memory you must unlink the 
      location and close any mmap file handles. You can use `cloudvolume.sharedmemory.unlink(...)`
      to help you unlink the shared memory file.

    EXPERT MODE WARNING: If you aren't sure you need this function (e.g. to relieve 
      memory pressure or improve performance in some way) you should use the ordinary 
      upload method of vol[:] = img. A typical use case is transferring arrays between 
      different processes without making copies. For reference, this feature was created
      for uploading a 62 GB array that originated in Julia.

    Required:
      location: (str) Shared memory location e.g. 'cloudvolume-shm-RANDOM-STRING'
        This typically corresponds to a file in `/dev/shm` or `/run/shm/`. It can 
        also be a file if you're using that for mmap.
      bbox: (Bbox or list of slices) the bounding box the shared array represents. For instance
        if you have a 1024x1024x128 volume and you're uploading only a 512x512x64 corner
        touching the origin, your Bbox would be `Bbox( (0,0,0), (512,512,64) )`.
    Optional:
      cutout_bbox: (bbox or list of slices) If you only want to upload a section of the
        array, give the bbox in volume coordinates (not image coordinates) that should 
        be cut out. For example, if you only want to upload 256x256x32 of the upper 
        rightmost corner of the above example but the entire 512x512x64 array is stored 
        in memory, you would provide: `Bbox( (256, 256, 32), (512, 512, 64) )`

        By default, just upload the entire image.

    Returns: void
    """
    bbox = Bbox.create(bbox)
    cutout_bbox = Bbox.create(cutout_bbox) if cutout_bbox else bbox.clone()

    if not bbox.contains_bbox(cutout_bbox):
      raise exceptions.AlignmentError("""
        The provided cutout is not wholly contained in the given array. 
        Bbox:        {}
        Cutout:      {}
      """.format(bbox, cutout_bbox))

    if self.autocrop:
      cutout_bbox = Bbox.intersection(cutout_bbox, self.bounds)

    if cutout_bbox.subvoxel():
      return

    shape = list(bbox.size3()) + [ self.num_channels ]
    mmap_handle, shared_image = sharedmemory.ndarray(
      location=location, shape=shape, 
      dtype=self.dtype, order=order, 
      readonly=True
    )

    delta_box = cutout_bbox.clone() - bbox.minpt
    cutout_image = shared_image[ delta_box.to_slices() ]
    
    self.image.upload(
      cutout_image, cutout_bbox.minpt, self.mip,
      parallel=self.parallel, 
      location=location, 
      location_bbox=bbox,
      order=order,
      use_shared_memory=True,
    )
    mmap_handle.close()

  def upload_from_file(self, location, bbox, order='F', cutout_bbox=None):
    """
    Upload from an mmapped file.

    tip: If you want to use slice notation, np.s_[...] will help in a pinch.

    Required:
      location: (str) Shared memory location e.g. 'cloudvolume-shm-RANDOM-STRING'
        This typically corresponds to a file in `/dev/shm` or `/run/shm/`. It can 
        also be a file if you're using that for mmap.
      bbox: (Bbox or list of slices) the bounding box the shared array represents. For instance
        if you have a 1024x1024x128 volume and you're uploading only a 512x512x64 corner
        touching the origin, your Bbox would be `Bbox( (0,0,0), (512,512,64) )`.
    Optional:
      cutout_bbox: (bbox or list of slices) If you only want to upload a section of the
        array, give the bbox in volume coordinates (not image coordinates) that should 
        be cut out. For example, if you only want to upload 256x256x32 of the upper 
        rightmost corner of the above example but the entire 512x512x64 array is stored 
        in memory, you would provide: `Bbox( (256, 256, 32), (512, 512, 64) )`

        By default, just upload the entire image.

    Returns: void
    """        
    bbox = Bbox.create(bbox)
    cutout_bbox = Bbox.create(cutout_bbox) if cutout_bbox else bbox.clone()

    if not bbox.contains_bbox(cutout_bbox):
      raise exceptions.AlignmentError("""
        The provided cutout is not wholly contained in the given array. 
        Bbox:        {}
        Cutout:      {}
      """.format(bbox, cutout_bbox))

    if self.autocrop:
      cutout_bbox = Bbox.intersection(cutout_bbox, self.bounds)

    if cutout_bbox.subvoxel():
      return

    shape = list(bbox.size3()) + [ self.num_channels ]
    mmap_handle, shared_image = sharedmemory.ndarray_fs(
      location=lib.toabs(location), shape=shape, 
      dtype=self.dtype, order=order, 
      readonly=True, lock=None
    )

    delta_box = cutout_bbox.clone() - bbox.minpt
    cutout_image = shared_image[ delta_box.to_slices() ]
    
    self.image.upload(
      cutout_image, cutout_bbox.minpt, self.mip,
      parallel=self.parallel, 
      location=lib.toabs(location), 
      location_bbox=bbox,
      order=order,
      use_file=True,
    )
    mmap_handle.close()

  def viewer(self, port=1337):
    import cloudvolume.server

    cloudvolume.server.view(self.cloudpath, port=port)

  def to_dask(self, chunks=None, name=None):
    """Return a dask array for this volume.

    Parameters
    ----------
    chunks: tuple of ints or tuples of ints
      Passed to ``da.from_array``, allows setting the chunks on
      initialisation, if the chunking scheme in the stored dataset is not
      optimal for the calculations to follow. Note that the chunking should
      be compatible with an underlying 4d array.
    name: str, optional
      An optional keyname for the array. Defaults to hashing the input

    Returns
    -------
    Dask array
    """
    import dask.array as da
    from dask.base import tokenize

    if chunks is None:
      chunks = tuple(self.chunk_size) + (self.num_channels, )
    if name is None:
      name = 'to-dask-' + tokenize(self, chunks)
    return da.from_array(self, chunks, name=name)



import os

try: 
  from http.server import BaseHTTPRequestHandler, HTTPServer
except ImportError:
  from SocketServer import TCPServer as HTTPServer
  from BaseHTTPServer import BaseHTTPRequestHandler

import json
from six.moves import range

import numpy as np
from tqdm import tqdm

from cloudfiles import CloudFiles

from cloudvolume.lib import Vec, Bbox, mkdir, save_images, yellow
from cloudvolume.paths import ExtractedPath

DEFAULT_PORT = 8080

def view(cloudpath, hostname="localhost", port=DEFAULT_PORT):
  """Start a local web app on the given port that lets you explore this cutout."""
  def handler(*args):
    return ViewerServerHandler(cloudpath, *args)

  myServer = HTTPServer((hostname, port), handler)
  print("Neuroglancer server listening to http://{}:{}".format(hostname, port))
  try:
    myServer.serve_forever()
  except KeyboardInterrupt:
    # extra \n to prevent display of "^CContinuing"
    print("\nContinuing program execution...")
  finally:
    myServer.server_close()

class ViewerServerHandler(BaseHTTPRequestHandler):
  def __init__(self, cloudpath, *args):
    self.cloudpath = cloudpath
    BaseHTTPRequestHandler.__init__(self, *args)

  def do_GET(self):  
    if self.path.find('..') != -1:
      self.send_error(403, "Relative paths are not allowed.")
      raise ValueError("Relative paths are not allowed.")

    path = self.path[1:]
    data = CloudFiles(self.cloudpath).get(path)

    if data is None:
      self.send_error(404, '/' + path + ": Not Found")
      return 

    self.send_response(200)
    self.serve_data(data)

  def serve_data(self, data):
    self.send_header('Content-type', 'application/octet-stream')
    self.send_header('Access-Control-Allow-Origin', '*')
    self.send_header('Content-length', str(len(data)))
    self.end_headers()
    self.wfile.write(data)



from intern.remote.boss import BossRemote
from intern.resource.boss.resource import ChannelResource, ExperimentResource, CoordinateFrameResource
from ...secrets import boss_credentials

from cloudvolume.datasource.precomputed.metadata import PrecomputedMetadata

class BossMetadata(PrecomputedMetadata):
  def __init__(self, cloudpath, cache, info=None):
    super(BossMetadata, self).__init__(
      cloudpath, cache, info=info, provenance=None
    )

  def commit_info(self):
    """BOSS doesn't support editing metadata after creation."""
    pass 

  def fetch_info(self):
    experiment = ExperimentResource(
      name=self.path.dataset, 
      collection_name=self.path.bucket
    )
    rmt = BossRemote(boss_credentials)
    experiment = rmt.get_project(experiment)

    coord_frame = CoordinateFrameResource(name=experiment.coord_frame)
    coord_frame = rmt.get_project(coord_frame)

    channel = ChannelResource(self.path.layer, self.path.bucket, self.path.dataset)
    channel = rmt.get_project(channel)    

    unit_factors = {
      'nanometers': 1,
      'micrometers': 1e3,
      'millimeters': 1e6,
      'centimeters': 1e7,
    }

    unit_factor = unit_factors[coord_frame.voxel_unit]

    cf = coord_frame
    resolution = [ cf.x_voxel_size, cf.y_voxel_size, cf.z_voxel_size ]
    resolution = [ int(round(_)) * unit_factor for _ in resolution ]

    bbox = Bbox(
      (cf.x_start, cf.y_start, cf.z_start),
      (cf.x_stop, cf.y_stop, cf.z_stop)
    )
    bbox.maxpt = bbox.maxpt 

    layer_type = 'unknown'
    if 'type' in channel.raw:
      layer_type = channel.raw['type']

    info = CloudVolume.create_new_info(
      num_channels=1,
      layer_type=layer_type,
      data_type=channel.datatype,
      encoding='raw',
      resolution=resolution,
      voxel_offset=bbox.minpt,
      volume_size=bbox.size3(),
      chunk_size=(512, 512, 16), # fixed in s3 implementation
    )

    each_factor = Vec(2,2,1)
    if experiment.hierarchy_method == 'isotropic':
      each_factor = Vec(2,2,2)
    
    factor = each_factor.clone()
    for _ in range(1, experiment.num_hierarchy_levels):
      self.add_scale(factor, info=info)
      factor *= each_factor

    return info

  def commit_provenance(self):
    """BOSS doesn't support provenance files."""
    pass

  def fetch_provenance(self):
    """BOSS doesn't support provenance files."""
    pass


import pytest

import os
import json 

from layer_harness import delete_layer
from cloudvolume.provenance import DataLayerProvenance, DatasetProvenance
from cloudvolume.storage import Storage

def test_dataset_provenance():
  fs = '/tmp/removeme/provenance/'
  delete_layer(fs)

  prov = DatasetProvenance()

  prov.dataset_name = 'ur-mom-2039'
  prov.dataset_description = 'EM serial section of your mom\'s brain'
  prov.organism = 'Male wild-type (C57BL/6) mouse'
  prov.imaged_date = 'March-Feb 2010'
  prov.imaged_by = 'gradstudent@princeton.edu'
  prov.references = [ 'doi:presigiousjournalofyourmom-12142' ]
  prov.owners = [ 'scientist@princeton.edu', 'techstaff@princeton.edu' ]

  with Storage('file://' + fs) as stor:
    stor.put_file('provenance', prov.serialize(), content_type='application/json')

  path = os.path.join(fs, 'provenance')

  with open(path, 'r') as f:
    data = json.loads(f.read())

  assert data == { 
    'dataset_name': 'ur-mom-2039',
    'dataset_description': 'EM serial section of your mom\'s brain',
    'organism': 'Male wild-type (C57BL/6) mouse',
    'imaged_date': 'March-Feb 2010',
    'imaged_by': 'gradstudent@princeton.edu',
    'references': [ 'doi:presigiousjournalofyourmom-12142' ],
    'owners': [ 'scientist@princeton.edu', 'techstaff@princeton.edu' ],
  }

  with Storage('file://' + fs) as stor:
    provjson = stor.get_file('provenance')
    provjson = provjson.decode('utf-8')
    prov = DatasetProvenance().from_json(provjson)

  assert prov.dataset_name == 'ur-mom-2039'
  assert prov.dataset_description == 'EM serial section of your mom\'s brain'
  assert prov.organism == 'Male wild-type (C57BL/6) mouse'
  assert prov.imaged_date == 'March-Feb 2010'
  assert prov.imaged_by == 'gradstudent@princeton.edu'
  assert prov.references == [ 'doi:presigiousjournalofyourmom-12142' ]
  assert prov.owners == [ 'scientist@princeton.edu', 'techstaff@princeton.edu' ]

def test_data_layer_provenance():
  fs = '/tmp/removeme/provenance/layer/'
  delete_layer(fs)

  prov = DataLayerProvenance()

  prov.description = 'example dataset'
  prov.sources = [ 'gs://seunglab-test/example/image' ]
  prov.processing = [ 
    { 'method': 'convnet', 'by': 'gradstudent@princeton.edu' },
  ]
  prov.owners = [ 'gradstudent@princeton.edu' ]

  with Storage('file://' + fs) as stor:
    stor.put_file('provenance', prov.serialize(), content_type='application/json')

  path = os.path.join(fs, 'provenance')

  with open(path, 'r') as f:
    data = json.loads(f.read())

  assert data == { 
    'description': 'example dataset', 
    'sources': [ 'gs://seunglab-test/example/image' ],
    'processing': [
      { 'method': 'convnet', 'by': 'gradstudent@princeton.edu' },
    ],
    'owners': [ 'gradstudent@princeton.edu' ]
  }

  with Storage('file://' + fs) as stor:
    provjson = stor.get_file('provenance')
    provjson = provjson.decode('utf-8')
    prov = DataLayerProvenance().from_json(provjson)

  assert prov.description == 'example dataset'
  assert prov.sources == [ 'gs://seunglab-test/example/image' ]
  assert prov.processing == [ { 'method': 'convnet', 'by': 'gradstudent@princeton.edu' } ]
  assert prov.owners == [ 'gradstudent@princeton.edu' ]




from .. import exceptions
from ..lib import yellow, Bbox, Vec

import numpy as np

NON_ALIGNED_WRITE = yellow(
  """
  Non-Aligned writes are disabled by default. There are several good reasons 
  not to use them. 

  1) Memory and Network Inefficiency
    Producing non-aligned writes requires downloading the chunks that overlap
    with the write area but are not wholly contained. They are then painted
    in the overlap region and added to the upload queue. This requires
    the outer shell of the aligned image to be downloaded painted, and uploaded. 
  2) Race Conditions
    If you are using multiple instances of CloudVolume, the partially overlapped 
    chunks will be downloaded, partially painted, and uploaded. If this procedure
    occurs close in time between two processes, the final chunk may be partially
    painted. If you are sure only one CloudVolume instance will be accessing an
    area at any one time, this is not a problem.

  If after reading this you are still sure you want non-aligned writes, you can
  set non_aligned_writes=True.

  Alignment Check: 
    Mip:             {mip}
    Chunk Size:      {chunk_size}
    Volume Offset:   {offset}
    Received:        {got} 
    Nearest Aligned: {check}
""")

class ImageSourceInterface(object):
  def download(self, bbox, mip):
    raise NotImplementedError()
  def upload(self, image, offset, mip):
    raise NotImplementedError()
  def exists(self, bbox, mip):
    raise NotImplementedError()
  def delete(self, bbox, mip):
    raise NotImplementedError()
  def transfer_to(self, cloudpath, bbox, mip):
    raise NotImplementedError()

def readonlyguard(fn):
  def guardfn(self, *args, **kwargs):
    if self.readonly:
      raise exceptions.ReadOnlyException(self.meta.cloudpath)
    return fn(self, *args, **kwargs)
  return guardfn

def check_grid_aligned(meta, img, bounds, mip, throw_error=False):
  """Raise a cloudvolume.exceptions.AlignmentError if the provided image is not grid aligned."""
  shape = Vec(*img.shape)[:3]
  alignment_check = bounds.expand_to_chunk_size(meta.chunk_size(mip), meta.voxel_offset(mip))
  alignment_check = Bbox.clamp(alignment_check, meta.bounds(mip))
  is_aligned = np.all(alignment_check.minpt == bounds.minpt) and np.all(alignment_check.maxpt == bounds.maxpt)
  
  if throw_error and is_aligned == False:
    msg = NON_ALIGNED_WRITE.format(
      mip=mip, chunk_size=meta.chunk_size(mip), 
      offset=meta.voxel_offset(mip), 
      got=bounds, check=alignment_check
    )
    raise exceptions.AlignmentError(msg)

  return is_aligned

def autocropfn(meta, image, bbox, mip):
  cropped_bbox = Bbox.intersection(bbox, meta.bounds(mip))
  dmin = np.abs(bbox.minpt - cropped_bbox.minpt)
  img_bbox = Bbox(dmin, dmin + cropped_bbox.size())
  image = image[ img_bbox.to_slices() ]

  return image, cropped_bbox





import pytest

import copy
import gzip
import json
import numpy as np
import os
import shutil
import sys

from cloudvolume import exceptions
from cloudvolume.exceptions import AlignmentError, ReadOnlyException
from cloudvolume import CloudVolume, chunks
from cloudvolume.lib import Bbox, Vec, yellow, mkdir
import cloudvolume.sharedmemory as shm
from layer_harness import (
  TEST_NUMBER,  
  delete_layer, create_layer
)
from cloudvolume.datasource.precomputed.common import cdn_cache_control
from cloudvolume.datasource.precomputed.image.tx import generate_chunks

def test_from_numpy():
  arr = np.random.randint(0, high=256, size=(128,128, 128))
  arr = np.asarray(arr, dtype=np.uint8)
  vol = CloudVolume.from_numpy(arr, max_mip=1)
  arr2 = vol[:,:,:]
  np.alltrue(arr == arr2)
  
  arr = np.random.randn(128,128, 128, 3)
  arr = np.asarray(arr, dtype=np.float32)
  vol = CloudVolume.from_numpy(arr, max_mip=1)
  arr2 = vol[:,:,:]
  np.alltrue(arr == arr2)
  shutil.rmtree('/tmp/image')

def test_cloud_access():
  CloudVolume('gs://seunglab-test/test_v0/image')
  CloudVolume('s3://seunglab-test/test_dataset/image')

def test_fill_missing():
  info = CloudVolume.create_new_info(
    num_channels=1, # Increase this number when we add more tests for RGB
    layer_type='image', 
    data_type='uint8', 
    encoding='raw',
    resolution=[ 1,1,1 ], 
    voxel_offset=[0,0,0], 
    volume_size=[128,128,64],
    mesh='mesh', 
    chunk_size=[ 64,64,64 ],
  )

  vol = CloudVolume('file:///tmp/cloudvolume/empty_volume', mip=0, info=info)
  vol.commit_info()

  vol = CloudVolume('file:///tmp/cloudvolume/empty_volume', mip=0, fill_missing=True)
  assert np.count_nonzero(vol[:]) == 0

  vol = CloudVolume('file:///tmp/cloudvolume/empty_volume', mip=0, fill_missing=True, cache=True)
  assert np.count_nonzero(vol[:]) == 0
  assert np.count_nonzero(vol[:]) == 0

  vol.cache.flush()
  delete_layer('/tmp/cloudvolume/empty_volume')

def test_has_data():
  delete_layer()
  cv, data = create_layer(size=(50,50,50,1), offset=(0,0,0))
  cv.add_scale((2,2,1))

  assert cv.image.has_data(0) == True
  assert cv.image.has_data(1) == False
  assert cv.image.has_data([1,1,1]) == True
  assert cv.image.has_data([2,2,1]) == False

  try:
    cv.image.has_data(2)
    assert False
  except exceptions.ScaleUnavailableError:
    pass

@pytest.mark.parametrize('green', (True, False))
def test_aligned_read(green):
  delete_layer()
  cv, data = create_layer(size=(50,50,50,1), offset=(0,0,0))
  cv.green_threads = green
  # the last dimension is the number of channels
  assert cv[0:50,0:50,0:50].shape == (50,50,50,1)
  assert np.all(cv[0:50,0:50,0:50] == data)
  
  delete_layer()
  cv, data = create_layer(size=(128,64,64,1), offset=(0,0,0))
  cv.green_threads = green
  # the last dimension is the number of channels
  assert cv[0:64,0:64,0:64].shape == (64,64,64,1) 
  assert np.all(cv[0:64,0:64,0:64] ==  data[:64,:64,:64,:])

  delete_layer()
  cv, data = create_layer(size=(128,64,64,1), offset=(10,20,0))
  cv.green_threads = green
  cutout = cv[10:74,20:84,0:64]
  # the last dimension is the number of channels
  assert cutout.shape == (64,64,64,1) 
  assert np.all(cutout == data[:64,:64,:64,:])
  # get the second chunk
  cutout2 = cv[74:138,20:84,0:64]
  assert cutout2.shape == (64,64,64,1) 
  assert np.all(cutout2 == data[64:128,:64,:64,:])

  assert cv[25, 25, 25].shape == (1,1,1,1)

def test_save_images():
  delete_layer()
  cv, data = create_layer(size=(50,50,50,1), offset=(0,0,0))  

  img = cv[:]
  directory = img.save_images()

  for z, fname in enumerate(sorted(os.listdir(directory))):
    assert fname == str(z).zfill(3) + '.png'

  shutil.rmtree(directory)

def test_bbox_read():
  delete_layer()
  cv, data = create_layer(size=(50,50,50,1), offset=(0,0,0))

  x = Bbox( (0,1,2), (48,49,50) )
  # the last dimension is the number of channels
  assert cv[x].shape == (48,48,48,1)
  assert np.all(cv[x] == data[0:48, 1:49, 2:50])  

def test_number_type_read():
  delete_layer()
  cv, data = create_layer(size=(50,50,50,1), offset=(0,0,0))

  for datatype in (
    np.uint8, np.uint16, np.uint32, np.uint64,
    np.int8, np.int16, np.int32, np.int64,
    np.float16, np.float32, np.float64,
    int, float
  ):
    x = datatype(5)

    # the last dimension is the number of channels
    assert cv[x,x,x].shape == (1,1,1,1)
    assert np.all(cv[x,x,x] == data[5,5,5])  

def test_ellipsis_read():
  delete_layer()
  cv, data = create_layer(size=(50,50,50,1), offset=(0,0,0))

  img = cv[...]
  assert np.all(img == data)

  img = cv[5:10, ...]
  assert np.all(img == data[5:10, :,:,:])

  img = cv[5:10, 7, 8, ...]
  assert np.all(np.squeeze(img) == data[5:10, 7, 8, 0])

  img = cv[5:10, ..., 8, 0]
  assert np.all(np.squeeze(img) == data[5:10, :, 8, 0])

  try:
    img = cv[..., 5, ..., 0]
  except ValueError:
    pass


def test_parallel_read():
  paths = [
    'gs://seunglab-test/test_v0/image',
    's3://seunglab-test/test_v0/image',
  ]

  for parallel in (2, True):
    for cloudpath in paths:
      vol1 = CloudVolume(cloudpath, parallel=1)
      vol2 = CloudVolume(cloudpath, parallel=parallel)

      data1 = vol1[:512,:512,:50]
      img = vol2[:512,:512,:50]
      assert np.all(data1 == vol2[:512,:512,:50])


def test_parallel_read_shm():
  paths = [
    'gs://seunglab-test/test_v0/image',
    's3://seunglab-test/test_v0/image',
  ]

  for cloudpath in paths:
    vol1 = CloudVolume(cloudpath, parallel=1)
    vol2 = CloudVolume(cloudpath, parallel=2)

    data1 = vol1[:512,:512,:50]
    data2 = vol2.download_to_shared_memory(np.s_[:512,:512,:50])
    assert np.all(data1 == data2)
    data2.close()
    vol2.unlink_shared_memory()

def test_parallel_write():
  delete_layer()
  cv, data = create_layer(size=(512,512,128,1), offset=(0,0,0))
  
  # aligned write
  cv.parallel = 2
  cv[:] = np.zeros(shape=(512,512,128,1), dtype=cv.dtype) + 5
  data = cv[:]
  assert np.all(data == 5)

  # non-aligned-write
  cv.parallel = 2
  cv.non_aligned_writes = True
  cv[1:,1:,1:] = np.zeros(shape=(511,511,127,1), dtype=cv.dtype) + 7
  data = cv[1:,1:,1:]
  assert np.all(data == 7)

  # thin non-aligned-write so that there's no aligned core
  cv.parallel = 2
  cv.non_aligned_writes = True
  cv[25:75,25:75,25:75] = np.zeros(shape=(50,50,50,1), dtype=cv.dtype) + 8
  data = cv[25:75,25:75,25:75]
  assert np.all(data == 8)


def test_parallel_shared_memory_write():
  delete_layer()
  cv, _ = create_layer(size=(256,256,128,1), offset=(0,0,0))

  shm_location = 'cloudvolume-test-shm-parallel-write'
  mmapfh, shareddata = shm.ndarray(shape=(256,256,128), dtype=np.uint8, location=shm_location)
  shareddata[:] = 1

  cv.parallel = 1
  cv.upload_from_shared_memory(shm_location, Bbox((0,0,0), (256,256,128)))
  assert np.all(cv[:] == 1)

  shareddata[:] = 2
  cv.parallel = 2
  cv.upload_from_shared_memory(shm_location, Bbox((0,0,0), (256,256,128)))
  assert np.all(cv[:] == 2)

  shareddata[:,:,:64] = 3
  cv.upload_from_shared_memory(shm_location, bbox=Bbox((0,0,0), (256,256,128)), 
    cutout_bbox=Bbox((0,0,0), (256,256,64)))
  assert np.all(cv[:,:,:64] == 3)    
  assert np.all(cv[:,:,64:128] == 2)    

  shareddata[:,:,:69] = 4
  cv.autocrop = True
  cv.upload_from_shared_memory(shm_location, bbox=Bbox((-5,-5,-5), (251,251,123)), 
    cutout_bbox=Bbox((-5,-5,-5), (128,128,64)))
  assert np.all(cv[:128,:128,:63] == 4)    
  assert np.all(cv[128:,128:,:64] == 3)    
  assert np.all(cv[:,:,64:128] == 2)    

  shareddata[:] = 0
  shareddata[:,0,0] = 1
  cv.upload_from_shared_memory(shm_location, bbox=Bbox((0,0,0), (256,256,128)), order='C')
  assert np.all(cv[0,0,:] == 1)
  assert np.all(cv[1,0,:] == 0)

  mmapfh.close()
  shm.unlink(shm_location)

def test_non_aligned_read():
  delete_layer()
  cv, data = create_layer(size=(128,64,64,1), offset=(0,0,0))

  # the last dimension is the number of channels
  assert cv[31:65,0:64,0:64].shape == (34,64,64,1) 
  assert np.all(cv[31:65,0:64,0:64] == data[31:65,:64,:64,:])

  # read a single pixel
  delete_layer()
  cv, data = create_layer(size=(64,64,64,1), offset=(0,0,0))
  # the last dimension is the number of channels
  assert cv[22:23,22:23,22:23].shape == (1,1,1,1) 
  assert np.all(cv[22:23,22:23,22:23] == data[22:23,22:23,22:23,:])

  # Test steps (negative steps are not supported)
  img1 = cv[::2, ::2, ::2, :]
  img2 = cv[:, :, :, :][::2, ::2, ::2, :]
  assert np.array_equal(img1, img2)

  # read a single pixel
  delete_layer()
  cv, data = create_layer(size=(256,256,64,1), offset=(3,7,11))
  # the last dimension is the number of channels
  assert cv[22:77:2, 22:197:3, 22:32].shape == (28,59,10,1) 
  assert data[19:74:2, 15:190:3, 11:21,:].shape == (28,59,10,1) 
  assert np.all(cv[22:77:2, 22:197:3, 22:32] == data[19:74:2, 15:190:3, 11:21,:])

def test_autocropped_read():
  delete_layer()
  cv, data = create_layer(size=(50,50,50,1), offset=(0,0,0))

  cv.autocrop = True
  cv.bounded = False

  # left overlap
  img = cv[-25:25,-25:25,-25:25]
  assert img.shape == (25,25,25,1)
  assert np.all(img == data[:25, :25, :25])

  # right overlap
  img = cv[40:60, 40:60, 40:60]
  assert img.shape == (10,10,10,1)
  assert np.all(img == data[40:, 40:, 40:])

  # containing
  img = cv[-100:100, -100:100, -100:100]
  assert img.shape == (50,50,50,1)
  assert np.all(img == data)

  # contained
  img = cv[10:20, 10:20, 10:20]
  assert img.shape == (10,10,10,1)
  assert np.all(img == data[10:20, 10:20, 10:20])

  # non-intersecting
  img = cv[100:120, 100:120, 100:120]
  assert img.shape == (0,0,0,1)
  assert np.all(img == data[0:0, 0:0, 0:0])    

@pytest.mark.parametrize('green', (True, False))
def test_download_upload_file(green):
  delete_layer()
  cv, _ = create_layer(size=(50,50,50,1), offset=(0,0,0))
  cv.green_threads = green

  mkdir('/tmp/file/')

  cv.download_to_file('/tmp/file/test', cv.bounds)
  cv2 = CloudVolume('file:///tmp/file/test2/', info=cv.info)
  cv2.upload_from_file('/tmp/file/test', cv.bounds)

  assert np.all(cv2[:] == cv[:])
  shutil.rmtree('/tmp/file/')

@pytest.mark.parametrize('green', (True, False))
def test_write(green):
  delete_layer()
  cv, _ = create_layer(size=(50,50,50,1), offset=(0,0,0))
  cv.green_threads = green

  replacement_data = np.zeros(shape=(50,50,50,1), dtype=np.uint8)
  cv[0:50,0:50,0:50] = replacement_data
  assert np.all(cv[0:50,0:50,0:50] == replacement_data)

  replacement_data = np.random.randint(255, size=(50,50,50,1), dtype=np.uint8)
  cv[0:50,0:50,0:50] = replacement_data
  assert np.all(cv[0:50,0:50,0:50] == replacement_data)

  replacement_data = np.random.randint(255, size=(50,50,50,1), dtype=np.uint8)
  bbx = Bbox((0,0,0), (50,50,50))
  cv[bbx] = replacement_data
  assert np.all(cv[bbx] == replacement_data)

  # out of bounds
  delete_layer()
  cv, _ = create_layer(size=(128,64,64,1), offset=(10,20,0))
  cv.green_threads = green
  with pytest.raises(ValueError):
    cv[74:150,20:84,0:64] = np.ones(shape=(64,64,64,1), dtype=np.uint8)
  
  # non-aligned writes
  delete_layer()
  cv, _ = create_layer(size=(128,64,64,1), offset=(10,20,0))
  cv.green_threads = green
  with pytest.raises(ValueError):
    cv[21:85,0:64,0:64] = np.ones(shape=(64,64,64,1), dtype=np.uint8)

  # test bounds check for short boundary chunk
  delete_layer()
  cv, _ = create_layer(size=(25,25,25,1), offset=(1,3,5))
  cv.green_threads = green
  cv.info['scales'][0]['chunk_sizes'] = [[ 11,11,11 ]]
  cv[:] = np.ones(shape=(25,25,25,1), dtype=np.uint8)

def test_non_aligned_write():
  delete_layer()
  offset = Vec(5,7,13)
  cv, _ = create_layer(size=(1024, 1024, 5, 1), offset=offset)

  cv[:] = np.zeros(shape=cv.shape, dtype=cv.dtype)

  # Write inside a single chunk

  onepx = Bbox( (10,200,15), (11,201,16) )
  try:
    cv[ onepx.to_slices() ] = np.ones(shape=onepx.size3(), dtype=cv.dtype)
    assert False
  except AlignmentError:
    pass
  
  cv.non_aligned_writes = True
  cv[ onepx.to_slices() ] = np.ones(shape=onepx.size3(), dtype=cv.dtype)
  answer = np.zeros(shape=cv.shape, dtype=cv.dtype)
  answer[ 5, 193, 2 ] = 1
  assert np.all(cv[:] == answer)

  # Write across multiple chunks
  cv[:] = np.zeros(shape=cv.shape, dtype=cv.dtype)
  cv.non_aligned_writes = True
  middle = Bbox( (512 - 10, 512 - 11, 0), (512 + 10, 512 + 11, 5) ) + offset
  cv[ middle.to_slices() ] = np.ones(shape=middle.size3(), dtype=cv.dtype)
  answer = np.zeros(shape=cv.shape, dtype=cv.dtype)
  answer[ 502:522, 501:523, : ] = 1
  assert np.all(cv[:] == answer)    

  cv.non_aligned_writes = False
  try:
    cv[ middle.to_slices() ] = np.ones(shape=middle.size3(), dtype=cv.dtype)
    assert False
  except AlignmentError:
    pass

  # Big inner shell
  delete_layer()
  cv, _ = create_layer(size=(1024, 1024, 5, 1), offset=offset)
  cv[:] = np.zeros(shape=cv.shape, dtype=cv.dtype)
  middle = Bbox( (512 - 150, 512 - 150, 0), (512 + 150, 512 + 150, 5) ) + offset

  try:
    cv[ middle.to_slices() ] = np.ones(shape=middle.size3(), dtype=cv.dtype)
    assert False
  except AlignmentError:
    pass

  cv.non_aligned_writes = True
  cv[ middle.to_slices() ] = np.ones(shape=middle.size3(), dtype=cv.dtype)
  answer = np.zeros(shape=cv.shape, dtype=cv.dtype)
  answer[ 362:662, 362:662, : ] = 1
  assert np.all(cv[:] == answer)    

from cloudvolume import view

def test_autocropped_write():
  delete_layer()
  cv, _ = create_layer(size=(100,100,100,1), offset=(0,0,0))

  cv.autocrop = True
  cv.bounded = False

  replacement_data = np.ones(shape=(300,300,300,1), dtype=np.uint8)
  cv[-100:200, -100:200, -100:200] = replacement_data
  assert np.all(cv[:,:,:] == replacement_data[0:100,0:100,0:100])
  
  replacement_data = np.random.randint(255, size=(100,100,100,1), dtype=np.uint8)
  
  cv[-50:50, -50:50, -50:50] = replacement_data
  assert np.all(cv[0:50,0:50,0:50] == replacement_data[50:, 50:, 50:])

  cv[50:150, 50:150, 50:150] = replacement_data
  assert np.all(cv[50:,50:,50:] == replacement_data[:50, :50, :50])

  cv[0:50, 0:50, 0:50] = replacement_data[:50,:50,:50]
  assert np.all(cv[0:50, 0:50, 0:50] == replacement_data[:50,:50,:50])    

  replacement_data = np.ones(shape=(100,100,100,1), dtype=np.uint8)
  cv[:] = replacement_data + 1
  cv[100:200, 100:200, 100:200] = replacement_data
  assert np.all(cv[:,:,:] != 1)

def test_writer_last_chunk_smaller():
  delete_layer()
  cv, data = create_layer(size=(100,64,64,1), offset=(0,0,0))
  cv.info['scales'][0]['chunk_sizes'] = [[ 64,64,64 ]]
  
  chunks = [ chunk for chunk in generate_chunks(cv.meta, data[:,:,:,:], (0,0,0), cv.mip) ]

  assert len(chunks) == 2

  startpt, endpt, spt, ept = chunks[0]
  assert np.array_equal(spt, (0,0,0))
  assert np.array_equal(ept, (64,64,64))
  assert np.all((endpt - startpt) == Vec(64,64,64))

  startpt, endpt, spt, ept = chunks[1]
  assert np.array_equal(spt, (64,0,0))
  assert np.array_equal(ept, (100,64,64))
  assert np.all((endpt - startpt) == Vec(36,64,64))

def test_write_compressed_segmentation():
  delete_layer()
  cv, data = create_layer(size=(128,64,64,1), offset=(0,0,0))

  cv.info['num_channels'] = 1
  cv.info['data_type'] = 'uint32'
  cv.scale['encoding'] = 'compressed_segmentation'
  cv.scale['compressed_segmentation_block_size'] = (8,8,8)
  cv.commit_info()

  cv[:] = data.astype(np.uint32)
  data2 = cv[:]

  assert np.all(data == data2)

  cv.info['data_type'] = 'uint64'
  cv.commit_info()

  cv[:] = data.astype(np.uint64)
  data2 = cv[:]
  
  assert np.all(data == data2)

# def test_reader_negative_indexing():
#     """negative indexing is supported"""
#     delete_layer()
#     cv, data = create_layer(size=(128,64,64,1), offset=(0,0,0))

#     # Test negative beginnings
#     img1 = cv[-1:, -1:, -1:, :]
#     img2 = cv[:, :, :, :][-1:, -1:, -1:, :]

#     assert np.array_equal(img1, img2)    

#     # Test negative ends
#     with pytest.raises(ValueError):
#         img1 = cv[::-1, ::-1, ::-1, :]

def test_negative_coords_upload_download():
  cv, data = create_layer(size=(128,64,64,1), offset=(-64,-64,-64))

  downloaded = cv[-64:64, -64:0, -64:0]
  assert np.all(data == downloaded)

def test_setitem_mismatch():
  delete_layer()
  cv, _ = create_layer(size=(64,64,64,1), offset=(0,0,0))

  with pytest.raises(ValueError):
    cv[0:64,0:64,0:64] = np.zeros(shape=(5,5,5,1), dtype=np.uint8)

def test_bounds():
  delete_layer()
  cv, _ = create_layer(size=(128,64,64,1), offset=(100,100,100))
  cv.bounded = True

  try:
    cutout = cv[0:,0:,0:,:]
    cutout = cv[100:229,100:165,100:165,0]
    cutout = cv[99:228,100:164,100:164,0]
  except ValueError:
    pass
  else:
    assert False

  # don't die
  cutout = cv[100:228,100:164,100:164,0]

  cv.bounded = False
  cutout = cv[0:,0:,0:,:]
  assert cutout.shape == (228, 164, 164, 1)

  assert np.count_nonzero(cutout) != 0

  cutout[100:,100:,100:,:] = 0

  assert np.count_nonzero(cutout) == 0

def test_provenance():
  delete_layer()
  cv, _ = create_layer(size=(64,64,64,1), offset=(0,0,0))

  provobj = json.loads(cv.provenance.serialize())
  provobj['processing'] = [] # from_numpy
  assert provobj == {
    "sources": [], 
    "owners": [], 
    "processing": [], 
    "description": ""
  }

  cv.provenance.sources.append('cooldude24@princeton.edu')
  cv.commit_provenance()
  cv.refresh_provenance()

  assert cv.provenance.sources == [ 'cooldude24@princeton.edu' ]

  # should not die
  cv = CloudVolume(cv.cloudpath, provenance={})
  cv = CloudVolume(cv.cloudpath, provenance={ 'sources': [] })
  cv = CloudVolume(cv.cloudpath, provenance={ 'owners': [] })
  cv = CloudVolume(cv.cloudpath, provenance={ 'processing': [] })
  cv = CloudVolume(cv.cloudpath, provenance={ 'description': '' })

  # should die
  try:
    cv = CloudVolume(cv.cloudpath, provenance={ 'sources': 3 })
    assert False
  except:
    pass

  cv = CloudVolume(cv.cloudpath, provenance="""{
    "sources": [ "wow" ]
  }""")

  assert cv.provenance.sources[0] == 'wow'

def test_info_provenance_cache():
  image = np.zeros(shape=(128,128,128,1), dtype=np.uint8)
  vol = CloudVolume.from_numpy(
    image, 
    voxel_offset=(0,0,0), 
    vol_path='gs://seunglab-test/cloudvolume/caching', 
    layer_type='image', 
    resolution=(1,1,1), 
    encoding='raw'
  )

  # Test Info
  vol.cache.enabled = True
  vol.cache.flush()
  info = vol.refresh_info()
  assert info is not None

  with open(os.path.join(vol.cache.path, 'info'), 'r') as infof:
    info = infof.read()
    info = json.loads(info)

  with open(os.path.join(vol.cache.path, 'info'), 'w') as infof:
    infof.write(json.dumps({ 'wow': 'amaze' }))

  info = vol.refresh_info()
  assert info == { 'wow': 'amaze' }
  vol.cache.enabled = False
  info = vol.refresh_info()
  assert info != { 'wow': 'amaze' }

  infopath = os.path.join(vol.cache.path, 'info')
  assert os.path.exists(infopath)

  vol.cache.flush_info()
  assert not os.path.exists(infopath)
  vol.cache.flush_info() # assert no error by double delete


  # Test Provenance
  vol.cache.enabled = True
  vol.cache.flush()
  prov = vol.refresh_provenance()
  assert prov is not None

  with open(os.path.join(vol.cache.path, 'provenance'), 'r') as provf:
    prov = provf.read()
    prov = json.loads(prov)

  with open(os.path.join(vol.cache.path, 'provenance'), 'w') as provf:
    prov['description'] = 'wow'
    provf.write(json.dumps(prov))

  prov = vol.refresh_provenance()
  assert prov['description'] == 'wow'
  vol.cache.enabled = False
  prov = vol.refresh_provenance()
  assert prov['description'] == ''

  provpath = os.path.join(vol.cache.path, 'provenance')
  vol.cache.flush_provenance()
  assert not os.path.exists(provpath)
  vol.cache.flush_provenance() # assert no error by double delete

def test_caching():
  image = np.zeros(shape=(128,128,128,1), dtype=np.uint8)
  image[0:64,0:64,0:64] = 1
  image[64:128,0:64,0:64] = 2
  image[0:64,64:128,0:64] = 3
  image[0:64,0:64,64:128] = 4
  image[64:128,64:128,0:64] = 5
  image[64:128,0:64,64:128] = 6
  image[0:64,64:128,64:128] = 7
  image[64:128,64:128,64:128] = 8

  dirpath = '/tmp/cloudvolume/caching-volume-' + str(TEST_NUMBER)
  layer_path = 'file://' + dirpath

  vol = CloudVolume.from_numpy(
    image, 
    voxel_offset=(0,0,0), 
    vol_path=layer_path, 
    layer_type='image', 
    resolution=(1,1,1), 
    encoding='raw',
    chunk_size=(64,64,64),
  )

  vol.cache.enabled = True
  vol.cache.flush()

  # Test that reading populates the cache
  read1 = vol[:,:,:]
  assert np.all(read1 == image)

  read2 = vol[:,:,:]
  assert np.all(read2 == image)

  assert len(vol.cache.list()) > 0

  files = vol.cache.list()
  validation_set = [
    '0-64_0-64_0-64',
    '64-128_0-64_0-64',
    '0-64_64-128_0-64',
    '0-64_0-64_64-128',
    '64-128_64-128_0-64',
    '64-128_0-64_64-128',
    '0-64_64-128_64-128',
    '64-128_64-128_64-128'
  ]
  assert set([ os.path.splitext(fname)[0] for fname in files ]) == set(validation_set)

  for i in range(8):
    fname = os.path.join(vol.cache.path, vol.key, validation_set[i]) + '.gz'
    with gzip.GzipFile(fname, mode='rb') as gfile:
      chunk = gfile.read()
    img3d = chunks.decode(
      chunk, 'raw', (64,64,64,1), np.uint8
    )
    assert np.all(img3d == (i+1))

  vol.cache.flush()
  assert not os.path.exists(vol.cache.path)

  # Test that writing populates the cache
  vol[:,:,:] = image

  assert os.path.exists(vol.cache.path)
  assert np.all(vol[:,:,:] == image)

  vol.cache.flush()

  # Test that partial reads work too
  result = vol[0:64,0:64,:]
  assert np.all(result == image[0:64,0:64,:])
  files = vol.cache.list()
  assert len(files) == 2
  result = vol[:,:,:]
  assert np.all(result == image)
  files = vol.cache.list()
  assert len(files) == 8

  vol.cache.flush()

  # Test Non-standard Cache Destination
  dirpath = '/tmp/cloudvolume/caching-cache-' + str(TEST_NUMBER)
  vol.cache.enabled = True
  vol.cache.path = dirpath
  vol[:,:,:] = image

  assert len(os.listdir(os.path.join(dirpath, vol.key))) == 8

  vol.cache.flush()

  # Test that caching doesn't occur when cache is not set
  vol.cache.enabled = False
  result = vol[:,:,:]
  if os.path.exists(vol.cache.path):
    files = vol.cache.list()
    assert len(files) == 0

  vol[:,:,:] = image
  if os.path.exists(vol.cache.path):
    files = vol.cache.list()
    assert len(files) == 0

  vol.cache.flush()

  # Test that deletion works too
  vol.cache.enabled = True
  vol[:,:,:] = image
  files = vol.cache.list()
  assert len(files) == 8
  vol.delete( np.s_[:,:,:] )
  files = vol.cache.list()
  assert len(files) == 0

  vol.cache.flush()    

  vol[:,:,:] = image
  files = vol.cache.list()
  assert len(files) == 8
  vol.cache.flush(preserve=np.s_[:,:,:])
  files = vol.cache.list()
  assert len(files) == 8
  vol.cache.flush(preserve=np.s_[:64,:64,:])
  files = vol.cache.list()
  assert len(files) == 2

  vol.cache.flush()

  vol[:,:,:] = image
  files = vol.cache.list()
  assert len(files) == 8
  vol.cache.flush_region(Bbox( (50, 50, 0), (100, 100, 10) ))
  files = vol.cache.list()
  assert len(files) == 4

  vol.cache.flush()

  vol[:,:,:] = image
  files = vol.cache.list()
  assert len(files) == 8
  vol.cache.flush_region(np.s_[50:100, 50:100, 0:10])
  files = vol.cache.list()
  assert len(files) == 4

  vol.cache.flush()

def test_cache_compression_setting():
  image = np.zeros(shape=(128,128,128,1), dtype=np.uint8)
  dirpath = '/tmp/cloudvolume/caching-validity-' + str(TEST_NUMBER)
  layer_path = 'file://' + dirpath

  vol = CloudVolume.from_numpy(
    image, 
    voxel_offset=(1,1,1), 
    vol_path=layer_path, 
    layer_type='image', 
    resolution=(1,1,1), 
    encoding='raw'
  )
  vol.cache.enabled = True
  vol.cache.flush()
  vol.commit_info()

  vol.cache.compress = None
  vol[:] = image
  assert all([ os.path.splitext(x)[1] == '.gz' for x in vol.cache.list() ])
  vol.cache.flush()

  vol.cache.compress = True
  vol[:] = image
  assert all([ os.path.splitext(x)[1] == '.gz' for x in vol.cache.list() ])
  vol.cache.flush()

  vol.cache.compress = False
  vol[:] = image
  assert all([ os.path.splitext(x)[1] == '' for x in vol.cache.list() ])
  vol.cache.flush()

  delete_layer(dirpath)

def test_cache_validity():
  image = np.zeros(shape=(128,128,128,1), dtype=np.uint8)
  dirpath = '/tmp/cloudvolume/caching-validity-' + str(TEST_NUMBER)
  layer_path = 'file://' + dirpath

  vol = CloudVolume.from_numpy(
    image, 
    voxel_offset=(1,1,1), 
    vol_path=layer_path, 
    layer_type='image', 
    resolution=(1,1,1), 
    encoding='raw'
  )
  vol.cache.enabled = True
  vol.cache.flush()
  vol.commit_info()

  def test_with_mock_cache_info(info, shoulderror):
    finfo = os.path.join(vol.cache.path, 'info')
    with open(finfo, 'w') as f:
      f.write(json.dumps(info))

    if shoulderror:
      try:
        CloudVolume(vol.cloudpath, cache=True)
      except ValueError:
        pass
      else:
        assert False
    else:
      CloudVolume(vol.cloudpath, cache=True)

  test_with_mock_cache_info(vol.info, shoulderror=False)

  info = vol.info.copy()
  info['scales'][0]['size'][0] = 666
  test_with_mock_cache_info(info, shoulderror=False)

  test_with_mock_cache_info({ 'zomg': 'wow' }, shoulderror=True)

  def tiny_change(key, val):
    info = vol.info.copy()
    info[key] = val
    test_with_mock_cache_info(info, shoulderror=True)

  tiny_change('type', 'zoolander')
  tiny_change('data_type', 'uint32')
  tiny_change('num_channels', 2)
  tiny_change('mesh', 'mesh')

  def scale_change(key, val, mip=0):
    info = vol.info.copy()
    info['scales'][mip][key] = val
    test_with_mock_cache_info(info, shoulderror=True)

  scale_change('voxel_offset', [ 1, 2, 3 ])
  scale_change('resolution', [ 1, 2, 3 ])
  scale_change('encoding', 'npz')

  vol.cache.flush()

  # Test no info file at all    
  CloudVolume(vol.cloudpath, cache=True)

  vol.cache.flush()

def test_pickling():
  import pickle
  delete_layer()
  cv, _ = create_layer(size=(128,64,64,1), offset=(0,0,0))

  pckl = pickle.dumps(cv)
  cv2 = pickle.loads(pckl)

  assert cv2.cloudpath == cv.cloudpath
  assert cv2.mip == cv.mip

def test_multiprocess():
  from concurrent.futures import ProcessPoolExecutor, as_completed

  delete_layer()
  cv, _ = create_layer(size=(128,64,64,1), offset=(0,0,0))
  cv.commit_info()

  # "The ProcessPoolExecutor class has known (unfixable) 
  # problems on Python 2 and should not be relied on 
  # for mission critical work."
  # https://pypi.org/project/futures/

  if sys.version_info[0] < 3:
    print(yellow("External multiprocessing not supported in Python 2."))
    return
  
  futures = []
  with ProcessPoolExecutor(max_workers=4) as ppe:
    for _ in range(0, 5):
      futures.append(ppe.submit(cv.refresh_info))

    for future in as_completed(futures):
      # an error should be re-raised in one of the futures
      future.result()

  delete_layer()

def test_exists():
  # Bbox version
  delete_layer()
  cv, _ = create_layer(size=(128,64,64,1), offset=(0,0,0))

  defexists = Bbox( (0,0,0), (128,64,64) )
  results = cv.exists(defexists)
  assert len(results) == 2
  assert results['1_1_1/0-64_0-64_0-64'] == True
  assert results['1_1_1/64-128_0-64_0-64'] == True

  fpath = os.path.join(cv.cloudpath, cv.key, '64-128_0-64_0-64')
  fpath = fpath.replace('file://', '') + '.gz'
  os.remove(fpath)

  results = cv.exists(defexists)
  assert len(results) == 2
  assert results['1_1_1/0-64_0-64_0-64'] == True
  assert results['1_1_1/64-128_0-64_0-64'] == False

  # Slice version
  delete_layer()
  cv, _ = create_layer(size=(128,64,64,1), offset=(0,0,0))

  defexists = np.s_[ 0:128, :, : ]

  results = cv.exists(defexists)
  assert len(results) == 2
  assert results['1_1_1/0-64_0-64_0-64'] == True
  assert results['1_1_1/64-128_0-64_0-64'] == True

  fpath = os.path.join(cv.cloudpath, cv.key, '64-128_0-64_0-64')
  fpath = fpath.replace('file://', '') + '.gz'
  os.remove(fpath)

  results = cv.exists(defexists)
  assert len(results) == 2
  assert results['1_1_1/0-64_0-64_0-64'] == True
  assert results['1_1_1/64-128_0-64_0-64'] == False

def test_delete():

  # Bbox version
  delete_layer()
  cv, _ = create_layer(size=(128,64,64,1), offset=(0,0,0))

  defexists = Bbox( (0,0,0), (128,64,64) )
  results = cv.exists(defexists)
  assert len(results) == 2
  assert results['1_1_1/0-64_0-64_0-64'] == True
  assert results['1_1_1/64-128_0-64_0-64'] == True


  cv.delete(defexists)
  results = cv.exists(defexists)
  assert len(results) == 2
  assert results['1_1_1/0-64_0-64_0-64'] == False
  assert results['1_1_1/64-128_0-64_0-64'] == False

  # Slice version
  delete_layer()
  cv, _ = create_layer(size=(128,64,64,1), offset=(0,0,0))

  defexists = np.s_[ 0:128, :, : ]

  results = cv.exists(defexists)
  assert len(results) == 2
  assert results['1_1_1/0-64_0-64_0-64'] == True
  assert results['1_1_1/64-128_0-64_0-64'] == True

  cv.delete(defexists)
  results = cv.exists(defexists)
  assert len(results) == 2
  assert results['1_1_1/0-64_0-64_0-64'] == False
  assert results['1_1_1/64-128_0-64_0-64'] == False

  # Check errors
  delete_layer()
  cv, _ = create_layer(size=(128,64,64,1), offset=(0,0,0))

  try:
    results = cv.exists( np.s_[1:129, :, :] )
    print(results)
  except exceptions.OutOfBoundsError:
    pass
  else:
    assert False

def test_delete_black_uploads():
  for parallel in (1,2):
    delete_layer()
    cv, _ = create_layer(size=(256,256,256,1), offset=(0,0,0))

    ls = os.listdir('/tmp/removeme/layer/1_1_1/')
    assert len(ls) == 64

    cv.parallel = parallel
    cv.delete_black_uploads = True
    cv[64:64+128,64:64+128,64:64+128] = 0

    ls = os.listdir('/tmp/removeme/layer/1_1_1/')
    assert len(ls) == (64 - 8) 

    cv[64:64+128,64:64+128,64:64+128] = 0

    ls = os.listdir('/tmp/removeme/layer/1_1_1/')
    assert len(ls) == (64 - 8) 

    cv.image.background_color = 1
    cv[:] = 1 
    ls = os.listdir('/tmp/removeme/layer/1_1_1/')
    assert len(ls) == 0

    cv[:] = 0
    ls = os.listdir('/tmp/removeme/layer/1_1_1/')
    assert len(ls) == 64


def test_transfer():
  # Bbox version
  delete_layer()
  cv, _ = create_layer(size=(128,64,64,1), offset=(0,0,0))

  cv.transfer_to('file:///tmp/removeme/transfer/', cv.bounds)

  ls = os.listdir('/tmp/removeme/transfer/1_1_1/')

  assert '0-64_0-64_0-64.gz' in ls
  assert len(ls) == 2

  assert os.path.exists('/tmp/removeme/transfer/info')
  assert os.path.exists('/tmp/removeme/transfer/provenance')

  dcv = CloudVolume("file:///tmp/removeme/transfer")
  dcv.info["dont_touch_me_bro"] = True
  dcv.commit_info()

  cv.transfer_to('file:///tmp/removeme/transfer/', cv.bounds)
  dcv.refresh_info()

  assert 'dont_touch_me_bro' in dcv.info

def test_cdn_cache_control():
  delete_layer()
  create_layer(size=(128,10,10,1), offset=(0,0,0))

  assert cdn_cache_control(None) == 'max-age=3600, s-max-age=3600'
  assert cdn_cache_control(0) == 'no-cache'
  assert cdn_cache_control(False) == 'no-cache'
  assert cdn_cache_control(True) == 'max-age=3600, s-max-age=3600'

  assert cdn_cache_control(1337) == 'max-age=1337, s-max-age=1337'
  assert cdn_cache_control('private, must-revalidate') == 'private, must-revalidate'

  try:
    cdn_cache_control(-1)
  except ValueError:
    pass
  else:
    assert False

def test_bbox_to_mip():
  info = {
    'data_type': 'uint8',
    'mesh': '',
    'num_channels': 1,
    'scales': [
      { 
        'chunk_sizes': [[64, 64, 1]],
        'encoding': 'raw',
        'key': '4_4_40',
        'resolution': [4, 4, 40],
        'size': [1024, 1024, 32],
        'voxel_offset': [35, 0, 1],
      },
      {
        'chunk_sizes': [[64, 64, 1]],
        'encoding': 'raw',
        'key': '8_8_40',
        'resolution': [8, 8, 40],
        'size': [512, 512, 32],
        'voxel_offset': [17, 0, 1],
      },
      {
        'chunk_sizes': [[64, 64, 1]],
        'encoding': 'raw',
        'key': '16_16_40',
        'resolution': [16, 16, 40],
        'size': [256, 256, 32],
        'voxel_offset': [8, 0, 1],
      },
      {
        'chunk_sizes': [[64, 64, 1]],
        'encoding': 'raw',
        'key': '32_32_40',
        'resolution': [32, 32, 40],
        'size': [128, 128, 32],
        'voxel_offset': [4, 0, 1],
      },
    ],
    'type': 'image'
  }
  
  cv = CloudVolume('file:///tmp/removeme/bbox_to_mip', info=info)

  bbox = Bbox( (35,0,1), (1024, 1024, 32))
  res = cv.bbox_to_mip(bbox, 0, 3)
  assert res.minpt.x == 4
  assert res.minpt.y == 0
  assert res.minpt.z == 1

  bbox = Bbox( (4, 0, 1), (128, 128, 32) )
  res = cv.bbox_to_mip(bbox, 3, 0)
  assert res.minpt.x == 32
  assert res.minpt.y == 0
  assert res.minpt.z == 1  

  res = cv.bbox_to_mip(bbox, 0, 0)
  assert res == bbox

def test_slices_from_global_coords():
  delete_layer()
  cv, _ = create_layer(size=(1024, 1024, 5, 1), offset=(7,0,0))

  scale = cv.info['scales'][0]
  scale = copy.deepcopy(scale)
  scale['voxel_offset'] = [ 3, 0, 0 ]
  scale['volume_size'] = [ 512, 512, 5 ]
  scale['resolution'] = [ 2, 2, 1 ]
  scale['key'] = '2_2_1'
  cv.info['scales'].append(scale)
  cv.commit_info()

  assert len(cv.available_mips) == 2

  cv.mip = 1
  slices = cv.slices_from_global_coords( Bbox( (100, 100, 1), (500, 512, 2) ) )
  result = Bbox.from_slices(slices)
  assert result == Bbox( (50, 50, 1), (250, 256, 2) )

  cv.mip = 0
  slices = cv.slices_from_global_coords( Bbox( (100, 100, 1), (500, 512, 2) ) )
  result = Bbox.from_slices(slices)
  assert result == Bbox( (100, 100, 1), (500, 512, 2) )

  slices = cv.slices_from_global_coords( np.s_[:,:,:] )
  result = Bbox.from_slices(slices)
  assert result == Bbox( (7, 0, 0), ( 1031, 1024, 5) )


def test_slices_to_global_coords():
  delete_layer()
  cv, _ = create_layer(size=(1024, 1024, 5, 1), offset=(7,0,0))

  scale = cv.info['scales'][0]
  scale = copy.deepcopy(scale)
  scale['voxel_offset'] = [ 3, 0, 0 ]
  scale['volume_size'] = [ 512, 512, 5 ]
  scale['resolution'] = [ 2, 2, 1 ]
  scale['key'] = '2_2_1'
  cv.info['scales'].append(scale)
  cv.commit_info()

  assert len(cv.available_mips) == 2

  cv.mip = 1
  slices = cv.slices_to_global_coords( Bbox( (100, 100, 1), (500, 512, 2) ) )

  result = Bbox.from_slices(slices)
  assert result == Bbox( (200, 200, 1), (1000, 1024, 2) )

  cv.mip = 0
  slices = cv.slices_to_global_coords( Bbox( (100, 100, 1), (500, 512, 2) ) )
  result = Bbox.from_slices(slices)
  assert result == Bbox( (100, 100, 1), (500, 512, 2) )

# def test_boss_download():
#   vol = CloudVolume('gs://seunglab-test/test_v0/image')
#   bossvol = CloudVolume('boss://automated_testing/test_v0/image')

#   vimg = vol[:,:,:5]
#   bimg = bossvol[:,:,:5]

#   assert np.all(bimg == vimg)
#   assert bimg.dtype == vimg.dtype

#   vol.bounded = False
#   vol.fill_missing = True
#   bossvol.bounded = False
#   bossvol.fill_missing = True

#   assert np.all(vol[-100:100,-100:100,-10:10] == bossvol[-100:100,-100:100,-10:10])

#   # BOSS using a different algorithm for creating downsamples
#   # so hard to compare 1:1 w/ pixels.
#   bossvol.bounded = True
#   bossvol.fill_missing = False
#   bossvol.mip = 1
#   bimg = bossvol[:,:,5:6]
#   assert np.any(bimg > 0)

def test_redirects():
  info = CloudVolume.create_new_info(
    num_channels=1, # Increase this number when we add more tests for RGB
    layer_type='image', 
    data_type='uint8', 
    encoding='raw',
    resolution=[ 1,1,1 ], 
    voxel_offset=[0,0,0], 
    volume_size=[128,128,64],
    mesh='mesh', 
    chunk_size=[ 64,64,64 ],
  )

  vol = CloudVolume('file:///tmp/cloudvolume/redirects_0', mip=0, info=info)
  vol.commit_info()
  vol.refresh_info()

  vol.info['redirect'] = 'file:///tmp/cloudvolume/redirects_0'
  vol.commit_info()
  vol.refresh_info()

  del vol.info['redirect']

  for i in range(0, 10):
    info['redirect'] = 'file:///tmp/cloudvolume/redirects_' + str(i + 1)  
    vol = CloudVolume('file:///tmp/cloudvolume/redirects_' + str(i), mip=0, info=info)
    vol.commit_info()
  else:
    del vol.info['redirect']
    vol.commit_info()

  vol = CloudVolume('file:///tmp/cloudvolume/redirects_0', mip=0)

  assert vol.cloudpath == 'file:///tmp/cloudvolume/redirects_9'

  info['redirect'] = 'file:///tmp/cloudvolume/redirects_10'
  vol = CloudVolume('file:///tmp/cloudvolume/redirects_9', mip=0, info=info)
  vol.commit_info()

  try:
    CloudVolume('file:///tmp/cloudvolume/redirects_0', mip=0)  
    assert False 
  except exceptions.TooManyRedirects:
    pass

  vol = CloudVolume('file:///tmp/cloudvolume/redirects_9', max_redirects=0)
  del vol.info['redirect']
  vol.commit_info()

  vol = CloudVolume('file:///tmp/cloudvolume/redirects_5', max_redirects=0)  
  vol.info['redirect'] = 'file:///tmp/cloudvolume/redirects_1'
  vol.commit_info()

  try:
    vol = CloudVolume('file:///tmp/cloudvolume/redirects_5')
    assert False
  except exceptions.CyclicRedirect:
    pass

  vol.info['redirect'] = 'file:///tmp/cloudvolume/redirects_6'
  vol.commit_info()

  vol = CloudVolume('file:///tmp/cloudvolume/redirects_1')

  try:
    vol[:,:,:] = 1
    assert False 
  except exceptions.ReadOnlyException:
    pass

  for i in range(0, 10):
    delete_layer('/tmp/cloudvolume/redirects_' + str(i))  
  
@pytest.mark.parametrize("compress_method", ('', None, True, False, 'gzip', 'br'))
def test_compression_methods(compress_method):
  path = "file:///tmp/cloudvolume/test" + '-' + str(TEST_NUMBER)


  # create a NG volume
  info = CloudVolume.create_new_info(
      num_channels=1,
      layer_type="image",
      data_type="uint16",  # Channel images might be 'uint8'
      encoding="raw",  # raw, jpeg, compressed_segmentation, fpzip, kempressed
      resolution=[4, 4, 40],  # Voxel scaling, units are in nanometers
      voxel_offset=[0, 0, 0],  # x,y,z offset in voxels from the origin
      chunk_size=[512, 512, 16],  # units are voxels
      volume_size=[512 * 4, 512 * 4, 16 * 40],  
  )
  vol = CloudVolume(path, info=info, compress=compress_method)
  vol.commit_info()

  image = np.random.randint(low=0, high=2 ** 16 - 1, size=(512, 512, 16, 1), dtype="uint16")

  vol[0:512, 0:512, 0:16] = image

  image_test = vol[0:512, 0:512, 0:16]
  
  delete_layer('/tmp/cloudvolume/test' + '-' + str(TEST_NUMBER))

  assert vol.compress == compress_method
  assert np.array_equal(image_test, image)

def test_mip_locking():
  delete_layer()
  cv, _ = create_layer(size=(1024, 1024, 2, 1), offset=(0,0,0))

  cv.meta.lock_mips(0)
  cv.meta.lock_mips([0])

  try:
    cv[:,:,:] = 0
    assert False 
  except ReadOnlyException:
    pass 

  cv.meta.unlock_mips(0)
  cv.meta.unlock_mips([0])

  cv[:,:,:] = 0

  try:
    cv.meta.lock_mips(1)
    assert False 
  except ValueError:
    pass 

  try:
    cv.meta.unlock_mips(1)
    assert False 
  except ValueError:
    pass 

  cv.add_scale((2,2,1))
  cv.commit_info()
  cv.mip = 1 
  cv[:] = 1

  cv.meta.lock_mips([0,1])

  try:
    cv[:,:,:] = 1
    assert False 
  except ReadOnlyException:
    pass 

  cv.mip = 0

  try:
    cv[:,:,:] = 1
    assert False 
  except ReadOnlyException:
    pass 



from cloudvolume import CloudVolume
import numpy as np
from copy import deepcopy

black = CloudVolume('gs://seunglab-test/test_v0/black')
info = deepcopy(black.info)

sizes = ( 
	# [128,128,1],
	# [128,128,2],
	# [128,128,4],
	# [128,128,4],
	# [128,128,8],
	# [128,128,16],
	# [128,128,32],
	# [128,128,64],
	[128,128,128],
	[1024,1024,1],
	[1024,1024,2],
	[1024,1024,4],
	[1024,1024,8],
	[1024,1024,16],
	[1024,1024,32],
	[1024,1024,64],
	[1024,1024,100], # max size of SNEMI3D
)

imgvol = CloudVolume('gs://seunglab-test/test_v0/black', progress=True)
img = imgvol[:]

for chunksize in sizes:
	info['scales'][0]['chunk_sizes'] = [ chunksize ]

	layer = 'black_{}_{}_{}'.format(*chunksize)

	vol = CloudVolume('gs://seunglab-test/test_v0/' + layer, info=info, compress='gzip', progress=True)
	vol.reset_scales()
	vol.commit_info()

	# vol[:] = np.zeros(shape=vol.shape, dtype=vol.dtype)
	vol[:] = img

	layer = 'black_uncompressed_{}_{}_{}'.format(*chunksize)

	vol = CloudVolume('gs://seunglab-test/test_v0/' + layer, info=info, compress='', progress=True)
	vol.reset_scales()
	vol.commit_info()

	vol[:] = img




class DoublyLinkedListIterator(object):
  def __init__(self, node, reverse=False):
    self.node = ListNode(None, node, node)
    self.reverse = reverse
  def next(self):
    return self.__next__()
  def __next__(self):
    if self.reverse:
      if self.node.prev is not None:
        self.node = self.node.prev
        return self.node
    else:
      if self.node.next is not None:
        self.node = self.node.next
        return self.node
    raise StopIteration()
  def __reversed__(self):
    return DoublyLinkedListIterator(self.node, not self.reverse)

class ListNode(object):
  def __init__(self, val, next, prev):
    self.val = val
    self.next = next
    self.prev = prev 

  def __iter__(self):
    return DoublyLinkedListIterator(self)

  def __reversed__(self):
    return DoublyLinkedListIterator(self, reverse=True)

  def __str__(self):
    return "ListNode({},{},{})".format(
      self.val, 
      self.next.val if self.next is not None else None,
      self.prev.val if self.prev is not None else None
    )

  def clone(self):
    return ListNode(self.val, self.next, self.prev)

class DoublyLinkedList(object):
  def __init__(self):
    self.head = None
    self.tail = None
    self.size = 0

  @classmethod
  def create(cls, lst):
    lst = iter(lst)

    dll = DoublyLinkedList()
    for val in lst:
      dll.append(val)

    return dll

  def __len__(self):
    return self.size

  def __iter__(self):
    return DoublyLinkedListIterator(self.head)

  def __reversed__(self):
    return DoublyLinkedListIterator(self.tail, reverse=True)

  def tolist(self):
    return [ x.val for x in self ]

  def is_empty(self):
    return self.head is None and self.tail is None

  def peek_head(self):
    if self.head is None:
      return None
    return self.head.val

  def peek_tail(self):
    if self.tail is None:
      return None
    return self.tail.val

  def promote_to_head(self, node):
    self.delete(node)

    if self.head is None:
      self.head = node
      node.next = None 
      node.prev = None
      self.tail = node
    else:
      head = self.head
      head.prev = node
      self.head = node
      node.prev = None 
      node.next = head

    self.size += 1

  def delete(self, node):
    nxt, prev = node.next, node.prev

    if self.head == node:
      self.head = nxt
    if self.tail == node:
      self.tail = prev

    if prev is not None:
      prev.next = nxt

    if nxt is not None:
      nxt.prev = prev

    self.size -= 1

    return node

  def delete_tail(self):
    if self.tail is None:
      return None
    elif self.tail == self.head:
      val = self.tail.val
      self.tail = None
      self.head = None
      self.size = 0
      return val

    node = self.tail
    self.tail = node.prev
    self.tail.next = None

    self.size -= 1

    return node.val

  def prepend(self, val):
    if self.head is None and self.tail is None:
      self.head = ListNode(val, None, None)
      self.tail = self.head
    elif self.head is None:
      self.head = ListNode(val, next=self.tail, prev=None)
      self.tail.prev = self.head
    else:
      prev_head = self.head
      self.head = ListNode(val, next=prev_head, prev=None)
      prev_head.prev = self.head

    self.size += 1

    return self

  def append(self, val):
    if self.head is None and self.tail is None:
      self.head = ListNode(val, None, None)
      self.tail = self.head
    elif self.tail is None:
      self.tail = ListNode(val, None, self.head)
      self.head.next = self.tail
    else:
      self.tail = ListNode(val, None, self.tail)
      self.tail.prev.next = self.tail

    self.size += 1

    return self

  def __str__(self):
    return str([ n.val for n in self ])

class LRU(object):
  def __init__(self, size=100):
    self.size = size
    self.queue = DoublyLinkedList()
    self.hash = {}

  def __len__(self):
    return self.queue.size

  def keys(self):
    return self.hash.keys()

  def values(self):
    return ( node.val for val in self.queue )

  def items(self):
    return ( (key, node.val) for key, node in self.hash.items() )

  def clear(self):
    self.queue = DoublyLinkedList()
    self.hash = {}

  def resize(self, new_size):
    if new_size < 0:
      raise ValueError("The LRU limit must be a positive number. Got: " + str(new_size))

    if new_size == 0:
      self.clear()
      return

    if new_size >= self.size:
      self.size = new_size
      return 

    while len(self.queue) > new_size:
      (key,val) = self.queue.delete_tail()
      del self.hash[key]

  def delete(self, key):
    if key not in self.hash:
      raise KeyError("{} not in cache.".format(key))

    node = self.hash[key]
    self.queue.delete(node)
    del self.hash[key]

  def get(self, key, default=None):
    if key not in self.hash:
      if default is None:
        raise KeyError("{} not in cache.".format(key))
      return default

    node = self.hash[key]
    self.queue.promote_to_head(node)

    return node.val[1]

  def set(self, key, val):
    if self.size == 0:
      return

    pair = (key,val)
    if key in self.hash:
      node = self.hash[key]
      node.val = pair
      self.queue.promote_to_head(node)
      return

    self.queue.prepend(pair)
    self.hash[key] = self.queue.head

    while len(self.queue) > self.size:
      (tkey,tval) = self.queue.delete_tail()
      del self.hash[tkey]     

  def __contains__(self, key):
    return key in self.hash

  def __getitem__(self, key):
    return self.get(key)

  def __setitem__(self, key, val):
    return self.set(key, val)

  def __delitem__(self, key):
    return self.delete(key)

  def __str__(self):
    return str(self.queue)






import pytest

import copy
import gzip
import json
import math
import numpy as np
import os
import shutil

from cloudvolume import CloudVolume, chunks, Storage, Skeleton
from cloudvolume.storage import SimpleStorage
from cloudvolume.lib import mkdir, Bbox, Vec, jsonify

from cloudvolume.datasource.precomputed.sharding import ShardingSpecification
from cloudvolume.exceptions import SkeletonDecodeError, SkeletonAttributeMixingError

info = CloudVolume.create_new_info(
  num_channels=1, # Increase this number when we add more tests for RGB
  layer_type='segmentation', 
  data_type='uint16', 
  encoding='raw',
  resolution=[1,1,1], 
  voxel_offset=(0,0,0), 
  skeletons=True,
  volume_size=(100, 100, 100),
  chunk_size=(64, 64, 64),
)

skel_info = {
  "@type": "neuroglancer_skeletons", 
  "transform": [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0], 
  "vertex_attributes": [
    {"id": "radius", "data_type": "float32", "num_components": 1}
  ], 
  "mip": 3,
}

def test_skeletons():
  
  # Skeleton of my initials
  # z=0: W ; z=1 S
  vertices = np.array([
    [ 0, 1, 0 ],
    [ 1, 0, 0 ],
    [ 1, 1, 0 ],

    [ 2, 0, 0 ],
    [ 2, 1, 0 ],
    [ 0, 0, 1 ],

    [ 1, 0, 1 ],
    [ 1, 1, 1 ],
    [ 0, 1, 1 ],

    [ 0, 2, 1 ],
    [ 1, 2, 1 ],
  ], np.float32)

  edges = np.array([
    [0, 1],
    [1, 2],
    [2, 3],
    [3, 4],
    [4, 5],
    [5, 6],
    [6, 7],
    [7, 8],
    [8, 9],
    [9, 10],
    [10, 11],
  ], dtype=np.uint32)

  radii = np.array([
    1.0,
    2.5,
    3.0,
    4.1,
    1.2,
    5.6,
    2000.123123,
    15.33332221,
    8128.124,
    -1,
    1824.03
  ], dtype=np.float32)

  vertex_types = np.array([
   1,
   2,
   3,
   5,
   8,
   2,
   0,
   5,
   9,
   11,
   22,
  ], dtype=np.uint8)

  vol = CloudVolume('file:///tmp/cloudvolume/test-skeletons', info=info)
  vol.skeleton.upload_raw(
    segid=1, vertices=vertices, edges=edges, 
    radii=radii, vertex_types=vertex_types
  )
  skel = vol.skeleton.get(1)

  assert skel.id == 1
  assert np.all(skel.vertices == vertices)
  assert np.all(skel.edges == edges)
  assert np.all(skel.radii == radii)
  assert np.all(skel.vertex_types == vertex_types)
  assert vol.skeleton.meta.skeleton_path == 'skeletons'
  assert not skel.empty()

  with SimpleStorage('file:///tmp/cloudvolume/test-skeletons/') as stor:
    rawskel = stor.get_file('skeletons/1')
    assert len(rawskel) == 8 + 11 * (12 + 8 + 4 + 1) 
    stor.delete_file('skeletons/1')
  
  try:
    vol.skeleton.get(5)
    assert False
  except SkeletonDecodeError:
    pass

def test_no_edges():
  vertices = np.array([
    [ 0, 1, 0 ],
    [ 1, 0, 0 ],
  ], np.float32)

  edges = None
  vol = CloudVolume('file:///tmp/cloudvolume/test-skeletons', info=info)
  vol.skeleton.upload_raw(2, vertices, edges)
  skel = vol.skeleton.get(2)

  assert skel.id == 2
  assert np.all(skel.vertices == vertices)
  assert np.all(skel.edges.shape == (0, 2))
  assert vol.skeleton.path == 'skeletons'
  assert skel.empty()

  with SimpleStorage('file:///tmp/cloudvolume/test-skeletons/') as stor:
    rawskel = stor.get_file('skeletons/2')
    assert len(rawskel) == 8 + 2 * (12 + 0 + 4 + 1) 
    stor.delete_file('skeletons/2')

def test_no_vertices():
  vertices = np.array([], np.float32).reshape(0,3)

  edges = None
  vol = CloudVolume('file:///tmp/cloudvolume/test-skeletons', info=info)
  vol.skeleton.upload_raw(3, vertices, edges)
  skel = vol.skeleton.get(3)

  assert skel.id == 3
  assert np.all(skel.vertices == vertices)
  assert np.all(skel.edges.shape == (0, 2))
  assert skel.empty()
  assert vol.skeleton.path == 'skeletons'

  with SimpleStorage('file:///tmp/cloudvolume/test-skeletons/') as stor:
    rawskel = stor.get_file('skeletons/3')
    assert len(rawskel) == 8 + 0 * (12 + 8 + 4 + 1)
    stor.delete_file('skeletons/3')

def test_consolidate():
  skel = Skeleton(
    vertices=np.array([
      (0, 0, 0),
      (1, 0, 0),
      (2, 0, 0),
      (0, 0, 0),
      (2, 1, 0),
      (2, 2, 0),
      (2, 2, 1),
      (2, 2, 2),
    ], dtype=np.float32),

    edges=np.array([
      [0, 1],
      [1, 2],
      [2, 3],
      [3, 4],
      [4, 5],
      [5, 6],
      [6, 7],
    ], dtype=np.uint32),

    radii=np.array([
      0, 1, 2, 3, 4, 5, 6, 7
    ], dtype=np.float32),

    vertex_types=np.array([
      0, 1, 2, 3, 4, 5, 6, 7
    ], dtype=np.uint8),
  )

  correct_skel = Skeleton(
    vertices=np.array([
      (0, 0, 0),
      (1, 0, 0),
      (2, 0, 0),
      (2, 1, 0),
      (2, 2, 0),
      (2, 2, 1),
      (2, 2, 2),
    ], dtype=np.float32),

    edges=np.array([
      [0, 1],
      [0, 2],
      [0, 3],
      [1, 2],
      [3, 4],
      [4, 5],
      [5, 6],
    ], dtype=np.uint32),

    radii=np.array([
      0, 1, 2, 4, 5, 6, 7
    ], dtype=np.float32),

    vertex_types=np.array([
      0, 1, 2, 4, 5, 6, 7
    ], dtype=np.uint8),
  )

  consolidated = skel.consolidate()

  assert np.all(consolidated.vertices == correct_skel.vertices)
  assert np.all(consolidated.edges == correct_skel.edges)
  assert np.all(consolidated.radii == correct_skel.radii)
  assert np.all(consolidated.vertex_types == correct_skel.vertex_types)

def test_remove_disconnected_vertices():
  skel = Skeleton(
    [ 
      (0,0,0), (1,0,0), (2,0,0),
      (0,1,0), (0,2,0), (0,3,0),
      (-1, -1, -1)
    ], 
    edges=[ 
      (0,1), (1,2), 
      (3,4), (4,5), (3,5)
    ],
    segid=666,
  )

  res = skel.remove_disconnected_vertices()
  assert res.vertices.shape[0] == 6
  assert res.edges.shape[0] == 5 
  assert res.radii.shape[0] == 6
  assert res.vertex_types.shape[0] == 6
  assert res.id == 666


def test_equivalent():
  assert Skeleton.equivalent(Skeleton(), Skeleton())

  identity = Skeleton([ (0,0,0), (1,0,0) ], [(0,1)] )
  assert Skeleton.equivalent(identity, identity)

  diffvertex = Skeleton([ (0,0,0), (0,1,0) ], [(0,1)])
  assert not Skeleton.equivalent(identity, diffvertex)

  single1 = Skeleton([ (0,0,0), (1,0,0) ], edges=[ (1,0) ])
  single2 = Skeleton([ (0,0,0), (1,0,0) ], edges=[ (0,1) ])
  assert Skeleton.equivalent(single1, single2)

  double1 = Skeleton([ (0,0,0), (1,0,0) ], edges=[ (1,0) ])
  double2 = Skeleton([ (0,0,0), (1,0,0) ], edges=[ (0,1) ])
  assert Skeleton.equivalent(double1, double2)

  double1 = Skeleton([ (0,0,0), (1,0,0), (1,1,0) ], edges=[ (1,0), (1,2) ])
  double2 = Skeleton([ (0,0,0), (1,0,0), (1,1,0) ], edges=[ (2,1), (0,1) ])
  assert Skeleton.equivalent(double1, double2)

  double1 = Skeleton([ (0,0,0), (1,0,0), (1,1,0), (1,1,3) ], edges=[ (1,0), (1,2), (1,3) ])
  double2 = Skeleton([ (0,0,0), (1,0,0), (1,1,0), (1,1,3) ], edges=[ (3,1), (2,1), (0,1) ])
  assert Skeleton.equivalent(double1, double2)

def test_cable_length():
  skel = Skeleton([ 
      (0,0,0), (1,0,0), (2,0,0), (3,0,0), (4,0,0), (5,0,0)
    ], 
    edges=[ (1,0), (1,2), (2,3), (3,4), (5,4) ],
    radii=[ 1, 2, 3, 4, 5, 6 ],
    vertex_types=[1, 2, 3, 4, 5, 6]
  )

  assert skel.cable_length() == (skel.vertices.shape[0] - 1)

  skel = Skeleton([ 
      (2,0,0), (1,0,0), (0,0,0), (0,5,0), (0,6,0), (0,7,0)
    ], 
    edges=[ (1,0), (1,2), (2,3), (3,4), (5,4) ],
    radii=[ 1, 2, 3, 4, 5, 6 ],
    vertex_types=[1, 2, 3, 4, 5, 6]
  )
  assert skel.cable_length() == 9

  skel = Skeleton([ 
      (1,1,1), (0,0,0), (1,0,0)
    ], 
    edges=[ (1,0), (1,2) ],
    radii=[ 1, 2, 3],
    vertex_types=[1, 2, 3]
  )
  assert abs(skel.cable_length() - (math.sqrt(3) + 1)) < 1e-6

def test_transform():
  skelv = Skeleton([ 
      (0,0,0), (1,0,0), (1,1,0), (1,1,3), (2,1,3), (2,2,3)
    ], 
    edges=[ (1,0), (1,2), (2,3), (3,4), (5,4) ],
    radii=[ 1, 2, 3, 4, 5, 6 ],
    vertex_types=[1, 2, 3, 4, 5, 6],
    segid=1337,
    transform=np.array([
      [2, 0, 0, 0],
      [0, 2, 0, 0],
      [0, 0, 2, 0],
    ])
  )

  skelp = skelv.physical_space()
  assert np.all(skelp.vertices == skelv.vertices * 2)
  assert np.all(skelv.vertices == skelp.voxel_space().vertices)

  skelv.transform = [
    [1, 0, 0, 1],
    [0, 1, 0, 2],
    [0, 0, 1, 3],
  ]

  skelp = skelv.physical_space()
  tmpskel = skelv.clone() 
  tmpskel.vertices[:,0] += 1
  tmpskel.vertices[:,1] += 2
  tmpskel.vertices[:,2] += 3
  assert np.all(skelp.vertices == tmpskel.vertices)
  assert np.all(skelp.voxel_space().vertices == skelv.vertices)


def test_downsample():
  skel = Skeleton([ 
      (0,0,0), (1,0,0), (1,1,0), (1,1,3), (2,1,3), (2,2,3)
    ], 
    edges=[ (1,0), (1,2), (2,3), (3,4), (5,4) ],
    radii=[ 1, 2, 3, 4, 5, 6 ],
    vertex_types=[1, 2, 3, 4, 5, 6],
    segid=1337,
  )

  def should_error(x):
    try:
      skel.downsample(x)
      assert False
    except ValueError:
      pass

  should_error(-1)
  should_error(0)
  should_error(.5)
  should_error(2.00000000000001)

  dskel = skel.downsample(1)
  assert Skeleton.equivalent(dskel, skel)
  assert dskel.id == skel.id
  assert dskel.id == 1337

  dskel = skel.downsample(2)
  dskel_gt = Skeleton(
    [ (0,0,0), (1,1,0), (2,1,3), (2,2,3) ], 
    edges=[ (1,0), (1,2), (2,3) ],
    radii=[1,3,5,6], vertex_types=[1,3,5,6] 
  )
  assert Skeleton.equivalent(dskel, dskel_gt)

  dskel = skel.downsample(3)
  dskel_gt = Skeleton(
    [ (0,0,0), (1,1,3), (2,2,3) ], edges=[ (1,0), (1,2) ],
    radii=[1,4,6], vertex_types=[1,4,6],
  )
  assert Skeleton.equivalent(dskel, dskel_gt)

  skel = Skeleton([ 
      (0,0,0), (1,0,0), (1,1,0), (1,1,3), (2,1,3), (2,2,3)
    ], 
    edges=[ (1,0), (1,2), (3,4), (5,4) ],
    radii=[ 1, 2, 3, 4, 5, 6 ],
    vertex_types=[1, 2, 3, 4, 5, 6]
  )
  dskel = skel.downsample(2)
  dskel_gt = Skeleton(
    [ (0,0,0), (1,1,0), (1,1,3), (2,2,3) ], 
    edges=[ (1,0), (2,3) ],
    radii=[1,3,4,6], vertex_types=[1,3,4,6] 
  )
  assert Skeleton.equivalent(dskel, dskel_gt)


def test_downsample_joints():
  skel = Skeleton([ 
      
                        (2, 3,0), # 0
                        (2, 2,0), # 1
                        (2, 1,0), # 2
      (0,0,0), (1,0,0), (2, 0,0), (3,0,0), (4,0,0), # 3, 4, 5, 6, 7
                        (2,-1,0), # 8
                        (2,-2,0), # 9
                        (2,-3,0), # 10

    ], 
    edges=[ 
                  (0, 1),
                  (1, 2),
                  (2, 5),
        (3,4), (4,5), (5, 6), (6,7),
                  (5, 8),
                  (8, 9),
                  (9,10)
    ],
    radii=[ 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 ],
    vertex_types=[ 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 ],
    segid=1337,
  )

  ds_skel = skel.downsample(2)
  ds_skel_gt = Skeleton([ 

                        (2, 3,0), # 0
                        
                        (2, 2,0), # 1
      (0,0,0),          (2, 0,0),     (4,0,0), # 2, 3, 4

                        (2,-2,0), # 5                        
                        (2,-3,0), # 6

    ], 
    edges=[ 
                  (0,1),
                  (1,3),
              (2,3),  (3,4), 
                  (3,5),
                  (5,6)
    ],
    radii=[ 0, 1, 3, 5, 7, 9, 10 ],
    vertex_types=[ 0, 1, 3, 5, 7, 9, 10 ],
    segid=1337,
  )

  assert Skeleton.equivalent(ds_skel, ds_skel_gt)


def test_read_swc():

  # From http://research.mssm.edu/cnic/swc.html
  test_file = """# ORIGINAL_SOURCE NeuronStudio 0.8.80
# CREATURE
# REGION
# FIELD/LAYER
# TYPE
# CONTRIBUTOR
# REFERENCE
# RAW
# EXTRAS
# SOMA_AREA
# SHINKAGE_CORRECTION 1.0 1.0 1.0
# VERSION_NUMBER 1.0
# VERSION_DATE 2007-07-24
# SCALE 1.0 1.0 1.0
1 1 14.566132 34.873772 7.857000 0.717830 -1
2 0 16.022520 33.760513 7.047000 0.463378 1
3 5 17.542000 32.604973 6.885001 0.638007 2
4 0 19.163984 32.022469 5.913000 0.602284 3
5 0 20.448090 30.822802 4.860000 0.436025 4
6 6 21.897903 28.881084 3.402000 0.471886 5
7 0 18.461960 30.289471 8.586000 0.447463 3
8 6 19.420759 28.730757 9.558000 0.496217 7"""

  skel = Skeleton.from_swc(test_file)
  assert skel.vertices.shape[0] == 8
  assert skel.edges.shape[0] == 7

  skel_gt = Skeleton(
    vertices=[
      [14.566132, 34.873772, 7.857000],
      [16.022520, 33.760513, 7.047000],
      [17.542000, 32.604973, 6.885001],
      [19.163984, 32.022469, 5.913000],
      [20.448090, 30.822802, 4.860000],
      [21.897903, 28.881084, 3.402000],
      [18.461960, 30.289471, 8.586000],
      [19.420759, 28.730757, 9.558000]
    ],
    edges=[ (0,1), (1,2), (2,3), (3,4), (4,5), (2,6), (7,6) ],
    radii=[ 
      0.717830, 0.463378, 0.638007, 0.602284, 
      0.436025, 0.471886, 0.447463, 0.496217
    ],
    vertex_types=[
      1, 0, 5, 0, 0, 6, 0, 6
    ],
  )

  assert Skeleton.equivalent(skel, skel_gt)

  skel = Skeleton.from_swc(skel.to_swc())
  assert np.all(np.abs(skel.vertices - skel_gt.vertices) < 0.00001)
  # sorts edges
  skel = skel.consolidate()
  skel_gt = skel_gt.consolidate()
  assert np.all(skel.edges == skel_gt.edges)
  assert np.all(np.abs(skel.radii - skel_gt.radii) < 0.00001)

  Nv = skel.vertices.shape[0]
  Ne = skel.edges.shape[0]

  for _ in range(10):
    skel = Skeleton.from_swc(skel.to_swc())
    assert skel.vertices.shape[0] == Nv 
    assert skel.edges.shape[0] == Ne


def test_components():
  skel = Skeleton(
    [ 
      (0,0,0), (1,0,0), (2,0,0),
      (0,1,0), (0,2,0), (0,3,0),
    ], 
    edges=[ 
      (0,1), (1,2), 
      (3,4), (4,5), (3,5)
    ],
    segid=666,
  )

  components = skel.components()
  assert len(components) == 2
  assert components[0].vertices.shape[0] == 3
  assert components[1].vertices.shape[0] == 3
  assert components[0].edges.shape[0] == 2
  assert components[1].edges.shape[0] == 3

  skel1_gt = Skeleton([(0,0,0), (1,0,0), (2,0,0)], [(0,1), (1,2)])
  skel2_gt = Skeleton([(0,1,0), (0,2,0), (0,3,0)], [(0,1), (0,2), (1,2)])

  assert Skeleton.equivalent(components[0], skel1_gt)
  assert Skeleton.equivalent(components[1], skel2_gt)

def test_caching():
  vol = CloudVolume('file:///tmp/cloudvolume/test-skeletons', 
    info=info, cache=True)

  vol.cache.flush()

  skel = Skeleton(
    [ 
      (0,0,0), (1,0,0), (2,0,0),
      (0,1,0), (0,2,0), (0,3,0),
    ], 
    edges=[ 
      (0,1), (1,2), 
      (3,4), (4,5), (3,5)
    ],
    segid=666,
  )

  vol.skeleton.upload(skel)

  assert vol.cache.list_skeletons() == [ '666.gz' ]

  skel.id = 1
  with open(os.path.join(vol.cache.path, 'skeletons/1'), 'wb') as f:
    f.write(skel.to_precomputed())

  cached_skel = vol.skeleton.get(1)

  assert cached_skel == skel

  vol.cache.flush()


def test_simple_merge():
  skel1 = Skeleton(
    [ (0,0,0), (1,0,0), (2,0,0),  ], 
    edges=[ (0,1), (1,2), ],
    segid=1,
  )

  skel2 = Skeleton(
    [ (0,0,1), (1,0,2), (2,0,3),  ], 
    edges=[ (0,1), (1,2), ],
    segid=1,
  )

  result = Skeleton.simple_merge([ skel1, skel2 ])

  expected = Skeleton(
    [ (0,0,0), (1,0,0), (2,0,0), (0,0,1), (1,0,2), (2,0,3), ], 
    edges=[ (0,1), (1,2), (3,4), (4,5) ],
    segid=1,
  )

  assert result == expected

  skel1.extra_attributes = [{
    "id": "wow",
    "data_type": "uint8",
    "components": 1,
  }]
  skel1.wow = np.array([1,2,3], dtype=np.uint8)

  skel2.extra_attributes = [{
    "id": "wow",
    "data_type": "uint8",
    "components": 1,
  }]
  skel2.wow = np.array([4,5,6], dtype=np.uint8)

  result = Skeleton.simple_merge([ skel1, skel2 ])
  expected.wow = np.array([1,2,3,4,5,6], dtype=np.uint8)

  assert result == expected

  skel2.extra_attributes[0]['data_type'] = np.uint8

  try:
    Skeleton.simple_merge([ skel1, skel2 ])
    assert False
  except SkeletonAttributeMixingError:
    pass

  skel2.extra_attributes[0]['data_type'] = 'uint8'
  skel2.extra_attributes.append({
    "id": "amaze",
    "data_type": "float32",
    "components": 2,
  })

  try:
    Skeleton.simple_merge([ skel1, skel2 ])
    assert False
  except SkeletonAttributeMixingError:
    pass

def test_sharded():
  skel = Skeleton(
    [ 
      (0,0,0), (1,0,0), (2,0,0),
      (0,1,0), (0,2,0), (0,3,0),
    ], 
    edges=[ 
      (0,1), (1,2), 
      (3,4), (4,5), (3,5)
    ],
    segid=1,
    extra_attributes=[
      {
          "id": "radius",
          "data_type": "float32",
          "num_components": 1,
      }
    ]
  ).physical_space()

  skels = {}
  for i in range(10):
    sk = skel.clone()
    sk.id = i
    skels[i] = sk.to_precomputed()

  mkdir('/tmp/removeme/skeletons/sharded/skeletons')
  with open('/tmp/removeme/skeletons/sharded/info', 'wt') as f:
    f.write(jsonify(info))

  for idxenc in ('raw', 'gzip'):
    for dataenc in ('raw', 'gzip'):

      spec = ShardingSpecification(
        'neuroglancer_uint64_sharded_v1', 
        preshift_bits=1,
        hash='murmurhash3_x86_128', 
        minishard_bits=2, 
        shard_bits=1, 
        minishard_index_encoding=idxenc, 
        data_encoding=dataenc,
      )
      skel_info['sharding'] = spec.to_dict()

      with open('/tmp/removeme/skeletons/sharded/skeletons/info', 'wt') as f:
        f.write(jsonify(skel_info))

      files = spec.synthesize_shards(skels)
      for fname in files.keys():
        with open('/tmp/removeme/skeletons/sharded/skeletons/' + fname, 'wb') as f:
          f.write(files[fname])      

      cv = CloudVolume('file:///tmp/removeme/skeletons/sharded/')
      assert cv.skeleton.meta.mip == 3

      for i in range(10):
        sk = cv.skeleton.get(i).physical_space()
        sk.id = 1
        assert sk == skel

      labels = []
      for fname in files.keys():
        lbls = cv.skeleton.reader.list_labels(fname, path='skeletons')
        labels += list(lbls)
      
      labels.sort()
      assert labels == list(range(10))

      for filename, shard in files.items():
        decoded_skels = cv.skeleton.reader.disassemble_shard(shard)
        for label, binary in decoded_skels.items():
          Skeleton.from_precomputed(binary)

      exists = cv.skeleton.reader.exists(list(range(11)), path='skeletons')
      assert exists == {
        0: 'skeletons/0.shard', 
        1: 'skeletons/0.shard', 
        2: 'skeletons/0.shard', 
        3: 'skeletons/0.shard', 
        4: 'skeletons/0.shard', 
        5: 'skeletons/0.shard', 
        6: 'skeletons/0.shard', 
        7: 'skeletons/0.shard', 
        8: 'skeletons/1.shard', 
        9: 'skeletons/1.shard',
        10: None,
      }

  shutil.rmtree('/tmp/removeme/skeletons')






import os
import sys

import numpy as np

import cloudvolume.sharedmemory as shm

def test_ndarray_fs():
	location = 'cloudvolume-shm-test-ndarray'
	array_like, array = shm.ndarray_fs(
		shape=(2,2,2), dtype=np.uint8, location=location, 
		lock=None, emulate_shm=True
	)
	assert np.all(array == np.zeros(shape=(2,2,2), dtype=np.uint8))
	array[:] = 100
	array_like.close()

	array_like, array = shm.ndarray_fs(
		shape=(2,2,2), dtype=np.uint8, location=location, 
		lock=None, emulate_shm=True
	)
	assert np.all(array[:] == 100)
	array_like.close()

	full_loc = os.path.join(shm.EMULATED_SHM_DIRECTORY, location)

	assert os.path.exists(full_loc)
	assert os.path.getsize(full_loc) == 8

	assert shm.unlink_fs(location) == True
	assert shm.unlink_fs(location) == False

	try:
		array_like, array = shm.ndarray_fs(
			shape=(2,2,2), dtype=np.uint8, location=location, 
			lock=None, readonly=True, emulate_shm=True
		)
		assert False
	except shm.SharedMemoryReadError:
		pass

	array_like, array = shm.ndarray_fs(
		shape=(2,2,2), dtype=np.uint8, location=location, 
		lock=None, emulate_shm=True
	)
	try:
		array_like, array = shm.ndarray_fs(
			shape=(200,200,200), dtype=np.uint8, location=location, 
			lock=None, readonly=True, emulate_shm=True
		)
		assert False
	except shm.SharedMemoryReadError:
		pass

	assert shm.unlink_fs(location) == True
	assert shm.unlink_fs(location) == False

	assert not os.path.exists(full_loc)

def test_ndarray_sh():
	# Don't bother testing on unsupported platforms.
	if shm.EMULATE_SHM:
		return

	import psutil

	location = 'cloudvolume-shm-test-ndarray'
	array_like, array = shm.ndarray_shm(shape=(2,2,2), dtype=np.uint8, location=location)
	assert np.all(array == np.zeros(shape=(2,2,2), dtype=np.uint8))
	array[:] = 100
	array_like.close()

	array_like, array = shm.ndarray_shm(shape=(2,2,2), dtype=np.uint8, location=location)
	assert np.all(array[:] == 100)
	array_like.close()

	filename = os.path.join(shm.SHM_DIRECTORY, location)

	assert os.path.exists(filename)
	assert os.path.getsize(filename) == 8

	assert shm.unlink_shm(location) == True
	assert shm.unlink_shm(location) == False

	assert not os.path.exists(filename)

	available = psutil.virtual_memory().available
	array_like, array = shm.ndarray_shm(shape=(available // 10,2,2), dtype=np.uint8, location=location)
	array_like.close()
	try:
		array_like, array = shm.ndarray_shm(shape=(available,2,2), dtype=np.uint8, location=location)
		assert False
	except shm.SharedMemoryAllocationError:
		pass

	assert shm.unlink_shm(location) == True
	assert shm.unlink_shm(location) == False

	try:
		array_like, array = shm.ndarray_shm(shape=(2,2,2), dtype=np.uint8, location=location, readonly=True)
		assert False
	except shm.SharedMemoryReadError:
		pass

	array_like, array = shm.ndarray_shm(shape=(2,2,2), dtype=np.uint8, location=location)
	try:
		array_like, array = shm.ndarray_shm(shape=(200,200,200), dtype=np.uint8, location=location, readonly=True)
		assert False
	except shm.SharedMemoryReadError:
		pass

	assert shm.unlink_shm(location) == True
	assert shm.unlink_shm(location) == False



from functools import partial
import os

import numpy as np
from six.moves import range
from tqdm import tqdm

from cloudfiles import CloudFiles, reset_connection_pools

from cloudvolume import lib, chunks
from cloudvolume.exceptions import AlignmentError
from cloudvolume.lib import ( 
  mkdir, clamp, xyzrange, Vec, 
  Bbox, min2, max2
)
from cloudvolume.scheduler import schedule_jobs
from cloudvolume.threaded_queue import DEFAULT_THREADS
from cloudvolume.volumecutout import VolumeCutout

import cloudvolume.sharedmemory as shm

from ... import check_grid_aligned
from .common import (
  fs_lock, parallel_execution, chunknames, 
  shade
) 
from ..common import (
  content_type, cdn_cache_control,
  should_compress
)
from .rx import download_chunks_threaded

def upload(
    meta, cache,
    image, offset, mip,
    compress=None,
    compress_level=None,
    cdn_cache=None,
    parallel=1,
    progress=False,
    delete_black_uploads=False, 
    background_color=0,
    non_aligned_writes=False,
    location=None, location_bbox=None, location_order='F',
    use_shared_memory=False, use_file=False,
    green=False, fill_missing=False
  ):
  """Upload img to vol with offset. This is the primary entry point for uploads."""

  if not np.issubdtype(image.dtype, np.dtype(meta.dtype).type):
    raise ValueError("""
      The uploaded image data type must match the volume data type. 

      Volume: {}
      Image: {}
      """.format(meta.dtype, image.dtype)
    )

  shape = Vec(*image.shape)[:3]
  offset = Vec(*offset)[:3]
  bounds = Bbox( offset, shape + offset)

  is_aligned = check_grid_aligned(
    meta, image, bounds, mip, 
    throw_error=(non_aligned_writes == False)
  )

  options = {
    "compress": compress,
    "compress_level": compress_level,
    "cdn_cache": cdn_cache,
    "parallel": parallel, 
    "progress": progress,
    "location": location, 
    "location_bbox": location_bbox,
    "location_order": location_order,
    "use_shared_memory": use_shared_memory,
    "use_file": use_file,
    "delete_black_uploads": delete_black_uploads,
    "background_color": background_color,
    "green": green,  
  }

  if is_aligned:
    upload_aligned(
      meta, cache, 
      image, offset, mip,
      **options
    )
    return

  # Upload the aligned core
  expanded = bounds.expand_to_chunk_size(meta.chunk_size(mip), meta.voxel_offset(mip))
  retracted = bounds.shrink_to_chunk_size(meta.chunk_size(mip), meta.voxel_offset(mip))
  core_bbox = retracted.clone() - bounds.minpt

  if not core_bbox.subvoxel():
    core_img = image[ core_bbox.to_slices() ] 
    upload_aligned(
      meta, cache, 
      core_img, retracted.minpt, mip,
      **options
    )

  # Download the shell, paint, and upload
  all_chunks = set(chunknames(expanded, meta.bounds(mip), meta.key(mip), meta.chunk_size(mip)))
  core_chunks = set(chunknames(retracted, meta.bounds(mip), meta.key(mip), meta.chunk_size(mip)))
  shell_chunks = all_chunks.difference(core_chunks)

  def shade_and_upload(img3d, bbox):
    # decode is returning non-writable chunk
    # we're throwing them away so safe to write
    img3d.setflags(write=1) 
    shade(img3d, bbox, image, bounds)
    threaded_upload_chunks(
      meta, cache, 
      img3d, mip,
      (( Vec(0,0,0), Vec(*img3d.shape[:3]), bbox.minpt, bbox.maxpt),), 
      compress=compress, cdn_cache=cdn_cache,
      progress=progress, n_threads=0, 
      delete_black_uploads=delete_black_uploads,
      green=green,
    )

  compress_cache = should_compress(meta.encoding(mip), compress, cache, iscache=True)

  download_chunks_threaded(
    meta, cache, mip, shell_chunks, fn=shade_and_upload,
    fill_missing=fill_missing, progress=progress, 
    compress_cache=compress_cache,
    green=green
  )

def upload_aligned(
    meta, cache,
    img, offset, mip,
    compress=None,
    compress_level=None,
    cdn_cache=None,
    progress=False,
    parallel=1, 
    location=None, 
    location_bbox=None, 
    location_order='F', 
    use_shared_memory=False,
    use_file=False,
    delete_black_uploads=False,
    background_color=0,
    green=False,
  ):
  global fs_lock

  chunk_ranges = list(generate_chunks(meta, img, offset, mip))

  if parallel == 1:
    threaded_upload_chunks(
      meta, cache, 
      img, mip, chunk_ranges, 
      progress=progress,
      compress=compress, cdn_cache=cdn_cache,
      delete_black_uploads=delete_black_uploads,
      background_color=background_color,
      green=green, compress_level=compress_level,
    )
    return

  length = (len(chunk_ranges) // parallel) or 1
  chunk_ranges_by_process = []
  for i in range(0, len(chunk_ranges), length):
    chunk_ranges_by_process.append(
      chunk_ranges[i:i+length]
    )

  # use_shared_memory means use a predetermined
  # shared memory location, not no shared memory 
  # at all.
  if not use_shared_memory:
    array_like, renderbuffer = shm.ndarray(
      shape=img.shape, dtype=img.dtype, 
      location=location, order=location_order, 
      lock=fs_lock
    )
    renderbuffer[:] = img

  cup = partial(child_upload_process, 
    meta, cache, 
    img.shape, offset, mip,
    compress, cdn_cache, progress,
    location, location_bbox, location_order, 
    delete_black_uploads, background_color, 
    green, compress_level=compress_level
  )

  parallel_execution(cup, chunk_ranges_by_process, parallel, cleanup_shm=location)

  # If manual mode is enabled, it's the 
  # responsibilty of the user to clean up
  if not use_shared_memory:
    array_like.close()
    shm.unlink(location)

def child_upload_process(
    meta, cache, 
    img_shape, offset, mip,
    compress, cdn_cache, progress,
    location, location_bbox, location_order, 
    delete_black_uploads, background_color,
    green, chunk_ranges, compress_level=None,
  ):
  global fs_lock
  reset_connection_pools()

  shared_shape = img_shape
  if location_bbox:
    shared_shape = list(location_bbox.size3()) + [ meta.num_channels ]

  array_like, renderbuffer = shm.ndarray(
    shape=shared_shape, 
    dtype=meta.dtype, 
    location=location, 
    order=location_order, 
    lock=fs_lock, 
    readonly=True
  )

  if location_bbox:
    cutout_bbox = Bbox( offset, offset + img_shape[:3] )
    delta_box = cutout_bbox.clone() - location_bbox.minpt
    renderbuffer = renderbuffer[ delta_box.to_slices() ]

  threaded_upload_chunks(
    meta, cache, 
    renderbuffer, mip, chunk_ranges, 
    compress=compress, cdn_cache=cdn_cache, progress=progress,
    delete_black_uploads=delete_black_uploads, 
    background_color=background_color,
    green=green, compress_level=compress_level,
  )
  array_like.close()

def threaded_upload_chunks(
    meta, cache, 
    img, mip, chunk_ranges, 
    compress, cdn_cache, progress,
    n_threads=DEFAULT_THREADS,
    delete_black_uploads=False,
    background_color=0,
    green=False,
    compress_level=None,
  ):
  
  if cache.enabled:
    mkdir(cache.path)

  while img.ndim < 4:
    img = img[ ..., np.newaxis ]

  remote = CloudFiles(meta.cloudpath, progress=progress)
  local = CloudFiles('file://' + cache.path, progress=progress)

  def do_upload(imgchunk, cloudpath):
    encoded = chunks.encode(imgchunk, meta.encoding(mip), meta.compressed_segmentation_block_size(mip))

    remote.put(
        path=cloudpath, 
        content=encoded,
        content_type=content_type(meta.encoding(mip)), 
        compress=should_compress(meta.encoding(mip), compress, cache),
        compression_level=compress_level,
        cache_control=cdn_cache_control(cdn_cache),
      )

    if cache.enabled:
        local.put(
          path=cloudpath,
          content=encoded, 
          content_type=content_type(meta.encoding(mip)), 
          compress=should_compress(meta.encoding(mip), compress, cache, iscache=True)
        )

  def do_delete(cloudpath):
    remote.delete(cloudpath)
    
    if cache.enabled:
      local.delete(cloudpath)

  def process(startpt, endpt, spt, ept):
    if np.array_equal(spt, ept):
      return

    imgchunk = img[ startpt.x:endpt.x, startpt.y:endpt.y, startpt.z:endpt.z, : ]

    # handle the edge of the dataset
    clamp_ept = min2(ept, meta.bounds(mip).maxpt)
    newept = clamp_ept - spt
    imgchunk = imgchunk[ :newept.x, :newept.y, :newept.z, : ]

    filename = "{}-{}_{}-{}_{}-{}".format(
      spt.x, clamp_ept.x,
      spt.y, clamp_ept.y, 
      spt.z, clamp_ept.z
    )

    cloudpath = meta.join(meta.key(mip), filename)

    if delete_black_uploads:
      if np.any(imgchunk != background_color):
        do_upload(imgchunk, cloudpath)
      else:
        do_delete(cloudpath)
    else:
      do_upload(imgchunk, cloudpath)

  schedule_jobs(
    fns=( partial(process, *vals) for vals in chunk_ranges ), 
    concurrency=n_threads, 
    progress=('Uploading' if progress else None),
    total=len(chunk_ranges),
    green=green,
  )

def generate_chunks(meta, img, offset, mip):
  shape = Vec(*img.shape)[:3]
  offset = Vec(*offset)[:3]

  bounds = Bbox( offset, shape + offset)

  alignment_check = bounds.round_to_chunk_size(meta.chunk_size(mip), meta.voxel_offset(mip))

  if not np.all(alignment_check.minpt == bounds.minpt):
    raise AlignmentError("""
      Only chunk aligned writes are supported by this function. 

      Got:             {}
      Volume Offset:   {} 
      Nearest Aligned: {}
    """.format(
      bounds, meta.voxel_offset(mip), alignment_check)
    )

  bounds = Bbox.clamp(bounds, meta.bounds(mip))

  img_offset = bounds.minpt - offset
  img_end = Vec.clamp(bounds.size3() + img_offset, Vec(0,0,0), shape)

  for startpt in xyzrange( img_offset, img_end, meta.chunk_size(mip) ):
    startpt = startpt.clone()
    endpt = min2(startpt + meta.chunk_size(mip), shape)
    spt = (startpt + bounds.minpt).astype(int)
    ept = (endpt + bounds.minpt).astype(int)
    yield (startpt, endpt, spt, ept)


from cloudvolume import CloudVolume, Skeleton
from cloudvolume.storage import SimpleStorage
from cloudvolume.datasource.precomputed.image.common import compressed_morton_code
from cloudvolume.datasource.precomputed.sharding import ShardingSpecification
from cloudvolume.exceptions import SpecViolation

from cloudvolume import Vec
from cloudvolume import exceptions
from layer_harness import delete_layer, create_layer

import numpy as np

def test_actual_example_hash():
  spec = ShardingSpecification.from_dict({
    "@type" : "neuroglancer_uint64_sharded_v1",
    "data_encoding" : "gzip",
    "hash" : "murmurhash3_x86_128",
    "minishard_bits" : 11,
    "minishard_index_encoding" : "gzip",
    "preshift_bits" : 6,
    "shard_bits" : 7
  })

  spec.validate()

  shard_loc = spec.compute_shard_location(1822975381)
  assert shard_loc.shard_number == '42'
  assert shard_loc.minishard_number == 18

  spec = ShardingSpecification.from_dict({
    "@type" : "neuroglancer_uint64_sharded_v1",
    "data_encoding" : "gzip",
    "hash" : "murmurhash3_x86_128",
    "minishard_index_encoding" : "gzip",
    "preshift_bits" : 2,
    "minishard_bits" : 3,
    "shard_bits" : 3
  })

  spec.hash = 'identity'
  label = 0b10101010
  shard_loc = spec.compute_shard_location(label)
  minishard_no = 0b010
  shard_no = 0b101

  assert shard_loc.minishard_number == minishard_no
  assert shard_loc.shard_number == str(shard_no)

def test_compressed_morton_code():
  cmc = lambda coord: compressed_morton_code(coord, grid_size=(3,3,3))

  assert cmc((0,0,0)) == 0b000000
  assert cmc((1,0,0)) == 0b000001
  assert cmc((2,0,0)) == 0b001000
  assert cmc((3,0,0)) == 0b001001
  assert cmc((2,2,0)) == 0b011000
  assert cmc((2,2,1)) == 0b011100

  cmc = lambda coord: compressed_morton_code(coord, grid_size=(2,3,1))

  assert cmc((0,0,0)) == 0b000000
  assert cmc((1,0,0)) == 0b000001
  assert cmc((0,0,7)) == 0b000100
  assert cmc((2,3,1)) == 0b011110

def test_image_sharding_hash():
  spec = ShardingSpecification(
    type="neuroglancer_uint64_sharded_v1",
    data_encoding="gzip",
    hash="identity",
    minishard_bits=6,
    minishard_index_encoding="gzip",
    preshift_bits=9,
    shard_bits=16,
  ) 

  point = Vec(144689, 52487, 2829)
  volume_size = Vec(*[248832, 134144, 7063])
  chunk_size = Vec(*[128, 128, 16])

  grid_size = np.ceil(volume_size / chunk_size).astype(np.uint32)
  gridpt = np.ceil(point / chunk_size).astype(np.int32)
  code = compressed_morton_code(gridpt, grid_size)
  loc = spec.compute_shard_location(code)

  assert loc.shard_number == '458d'



def test_sharding_spec_validation():
  spec = ShardingSpecification(
    type="neuroglancer_uint64_sharded_v1",
    data_encoding="gzip",
    hash="murmurhash3_x86_128",
    minishard_bits=11,
    minishard_index_encoding="gzip",
    preshift_bits=6,
    shard_bits=7,
  ) 

  spec.validate()
  
  spec.minishard_bits = 0
  spec.shard_bits = 0
  spec.validate()
  
  spec.minishard_bits = 64
  spec.shard_bits = 0
  spec.validate()

  spec.minishard_bits = 0
  spec.shard_bits = 64
  spec.validate()

  spec.minishard_bits = 1
  spec.shard_bits = 64
  try:
    spec.validate()
    assert False
  except SpecViolation:
    pass

  spec.minishard_bits = 64
  spec.shard_bits = 1
  try:
    spec.validate()
    assert False
  except SpecViolation:
    pass

  spec.minishard_bits = 11
  spec.shard_bits = 7
  
  spec.hash = 'identity'
  spec.hash = 'murmurhash3_x86_128'
  try:
    spec.hash = 'murmurhash3_X86_128'
    assert False 
  except SpecViolation:
    pass

  try:
    spec.hash = 'something else'
    assert False
  except SpecViolation:
    pass

  try:
    spec.hash = ''
  except SpecViolation:
    pass

  spec = ShardingSpecification(
    type="neuroglancer_uint64_sharded_v1",
    data_encoding="gzip",
    hash="murmurhash3_x86_128",
    minishard_bits=11,
    minishard_index_encoding="gzip",
    preshift_bits=6,
    shard_bits=7,
  ) 

  spec.preshift_bits = 0
  spec.validate()

  spec.preshift_bits = 63
  spec.validate()

  spec.preshift_bits = 32
  spec.validate()

  try:
    spec.preshift_bits = 64
    spec.validate()
    assert False
  except SpecViolation:
    pass

  try:
    spec.preshift_bits = -1
    spec.validate()
    assert False
  except SpecViolation:
    pass

  spec.preshift_bits = 5

  spec.minishard_index_encoding = 'raw'
  spec.validate()

  spec.minishard_index_encoding = 'gzip'
  spec.validate()

  spec.data_encoding = 'raw'
  spec.validate()

  spec.data_encoding = 'gzip'
  spec.validate()

  try:
    spec.type = 'lol_my_type'
    spec.validate()
    assert False
  except SpecViolation:
    pass

def test_skeleton_fidelity():
  segid = 1822975381
  cv = CloudVolume('gs://seunglab-test/sharded')
  sharded_skel = cv.skeleton.get(segid)

  with SimpleStorage('gs://seunglab-test/sharded') as stor:
    binary = stor.get_file('skeletons/' + str(segid))

  unsharded_skel = Skeleton.from_precomputed(binary, 
    segid=1822975381, vertex_attributes=cv.skeleton.meta.info['vertex_attributes']
  )

  assert sharded_skel == unsharded_skel

def test_image_fidelity():
  point = (142195, 64376, 3130)
  cv = CloudVolume('gs://seunglab-test/sharded')
  img = cv.download_point(point, mip=0, size=128)

  N_labels = np.unique(img).shape[0]

  assert N_labels == 144

def test_write_image_shard():
  delete_layer()
  cv, data = create_layer(size=(256,256,256,1), offset=(0,0,0))

  spec = {
    "@type" : "neuroglancer_uint64_sharded_v1",
    "data_encoding" : "gzip",
    "hash" : "murmurhash3_x86_128",
    "minishard_bits" : 1,
    "minishard_index_encoding" : "raw",
    "preshift_bits" : 3,
    "shard_bits" : 0
  }
  cv.scale['sharding'] = spec

  cv[:] = data
  sharded_data = cv[:]
  assert np.all(data == sharded_data)

  spec['shard_bits'] = 1
  try:
    cv[:] = data
    assert False
  except exceptions.AlignmentError:
    pass


